# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations regarding drawing on images"""

import colorsys
import copy
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageFont

from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.segmentation import PolygonType, Segmentation


def generate_detector_colors(num_classes: int):  # type: ignore
    """
    Generates a color palette for object detector

    Args:
        num_classes: int, the number of classes to distinguish

    Returns:
        List[Tuples(int, int, int)] specifying bounding box color
        with len(List) == number of classes

    """

    # Generate colors for drawing bounding boxes.
    hsv_tuples = [(x / num_classes, 1.0, 1.0) for x in range(num_classes)]

    hsv_colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))

    rgb_colors = list(
        map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), hsv_colors)
    )

    np.random.seed(10101)  # Fixed seed for consistent colors across runs.
    np.random.shuffle(rgb_colors)  # Shuffle colors to decorrelate adjacent classes.
    np.random.seed(None)  # Reset seed to default.

    return rgb_colors


def draw_on_image(
    frame: np.ndarray,  # type: ignore[type-arg]
    rgb_colors: List[Tuple[int, int, int]],
    cv2_font: int = cv2.FONT_HERSHEY_COMPLEX,
    bounding_boxes: Optional[List[BoundingBox]] = None,
    segmentations: Optional[List[Segmentation]] = None,
    flip_image: bool = False,
    draw_caption: Optional[bool] = True,
    thickness: Optional[int] = None,
    font_scale: float = 0.9,
) -> np.ndarray:  # type: ignore[type-arg]
    """
    Draws annotations on the given image
    Args:
        frame: Numpy array, the image
        rgb_colors: List of colors, for matching a class-name to an index of the rgb-color list
        cv2_font: int, font of cv2 textual annotations
        bounding_boxes: Optional List of BoundingBox objects
        segmentations: Optional List of Segmentation objects
        flip_image: Bool, whether to flip the image or not
        draw_caption: Optional bool, whether to draw a caption text or not
        thickness: Optional int, defines the thickness of the annotation lines
        font_scale: float, defines the size of a font

    Returns: Numpy array, the given image decorated with annotations

    """

    frame = copy.deepcopy(frame)

    if bounding_boxes is not None:
        for bounding_box in bounding_boxes:
            frame = draw_bbox_cv2(
                frame=frame,
                color=rgb_colors[bounding_box.class_id],
                box=bounding_box.box,
                flip_image=flip_image,
                cv2_font=cv2_font,
                label=bounding_box.class_name,
                draw_caption=draw_caption,
                thickness=thickness,
                font_scale=font_scale,
                score=bounding_box.score,
            )

    if segmentations is not None:
        for segmentation in segmentations:
            frame = draw_polygon_cv2(
                frame=frame,
                polygon=segmentation.polygon,
                color=rgb_colors[segmentation.class_id],
                flip_image=flip_image,
                cv2_font=cv2_font,
                label=segmentation.class_name,
                draw_caption=draw_caption,
                thickness=thickness,
                font_scale=font_scale,
                score=segmentation.score,
            )

    return frame


def draw_polygon_cv2(
    frame: np.ndarray,  # type: ignore[type-arg]
    color: Tuple[int, int, int],
    polygon: Optional[PolygonType],
    flip_image: bool = False,
    cv2_font: int = cv2.FONT_HERSHEY_COMPLEX,
    label: Optional[str] = None,
    draw_caption: Optional[bool] = True,
    thickness: Optional[int] = None,
    font_scale: float = 0.9,
    score: Optional[float] = None,
) -> np.ndarray:  # type: ignore[type-arg]
    """
    Draw a polygon on a given image
    Args:
        frame: Numpy array, the image
        color: Tuple of three int values, a rgb color
        polygon: Optional List of Tuples, the polygon coordinates # TODO why Optional?
        flip_image: Bool, whether to flip the image or not
        cv2_font: int, font of cv2 textual annotations
        label: Optional string, label (class information) of polygon
        draw_caption: Optional bool, whether to draw a caption text or not
        thickness: Optional int, defines the thickness of the polygon lines
        font_scale: float, defines the size of a font
        score: Optional float, defines the confidence for the label of the polygon

    Returns: Numpy array, the given image decorated with a polygon

    """

    if polygon is None:
        return frame

    if thickness is None:
        thickness = (frame.shape[0] + frame.shape[1]) // 450

    # TODO: add box.to_tuple(...) ?
    if flip_image:
        # TODO: what to do when image is flipped?
        pass
    else:
        pts = np.array(polygon, np.int32)
        pts = pts.reshape((-1, 1, 2))

        frame = cv2.polylines(
            img=frame, pts=[pts], isClosed=True, color=color, thickness=thickness
        )

    # TODO: add caption
    if draw_caption and label is not None:
        if score is not None:
            caption = f"{label}_{score:.2f}"
        else:
            caption = label

        cv2.putText(
            img=frame,
            text=caption,
            org=(int(polygon[0][0]), int(polygon[0][1])),
            fontFace=cv2_font,
            fontScale=font_scale,
            color=color,
            thickness=thickness,
        )

    return frame


def draw_bbox_cv2(
    frame: np.ndarray,  # type: ignore[type-arg]
    color: Tuple[int, int, int],
    box: Box,
    flip_image: bool = False,
    cv2_font: int = cv2.FONT_HERSHEY_COMPLEX,
    label: Optional[str] = None,
    draw_caption: Optional[bool] = True,
    thickness: Optional[int] = None,
    font_scale: float = 0.9,
    score: Optional[float] = None,
    text_origin: Optional[Tuple[int, int]] = None,
) -> np.ndarray:  # type: ignore[type-arg]
    """
    Draws a bounding box on the given image
    Args:
        frame: Numpy array, the image
        color: Tuple of three int values, a rgb color
        box: Box, a bounding box object
        flip_image: Bool, whether to flip the image or not
        cv2_font: int, font of cv2 textual annotations
        label: Optional string, label (class information) of polygon
        draw_caption: Optional bool, whether to draw a caption text or not
        thickness: Optional int, defines the thickness of the box lines
        font_scale: float, defines the size of a font
        score: Optional float, defines the confidence for the label of the polygon
        text_origin: position of text in image

    Returns: Numpy array, the given image decorated with a bounding box

    """

    if box is None:
        return frame

    img_w = frame.shape[0]

    if thickness is None:
        thickness = (frame.shape[0] + frame.shape[1]) // 450

    # TODO: add box.to_tuple(...) ?
    if flip_image:
        frame = cv2.rectangle(
            frame,
            (img_w - box.xmin, box.ymin),
            (img_w - box.xmax, box.xmin),
            color,
            thickness,
        )
    else:
        frame = cv2.rectangle(frame, (box.xmin, box.ymin), (box.xmax, box.ymax), color, thickness)

    if draw_caption and label is not None:
        # TODO: adapt position?
        # if box.ymin - label_size[1] >= 0:
        #     if flip_image:
        #         text_origin = (img_w - box.xmin, box.ymin - label_size[1])
        #     else:
        #         text_origin = (box.xmin, box.ymin - label_size[1])
        # else:
        #     if flip_image:
        #         text_origin = (img_w - box.xmin, box.ymin + 1)
        #     else:
        #         pass
        #
        # label_size = cv2.getTextSize(label, cv2_font, font_scale, thickness)

        if text_origin is None:
            text_x = box.xmin
            text_y = max(box.ymin - 5, 20)
            text_origin = (text_x, text_y)

        if score is not None:
            caption = f"{label}_{score:.2f}"
        else:
            caption = label

        cv2.putText(
            img=frame,
            text=caption,
            org=text_origin,
            fontFace=cv2_font,
            fontScale=font_scale,
            color=color,
            thickness=thickness,
        )

    return frame


def draw_on_pil_image(
    image: Union[np.ndarray, Image.Image],  # type: ignore[type-arg]
    bounding_boxes: List[BoundingBox],
    font_path: str,
    rgb_colors: List[Tuple[int, int, int]],
    thickness: int = 2,
    fill_background: bool = True,
    is_ground_truth: bool = False,
) -> np.ndarray:  # type: ignore[type-arg]
    """
    Draw a given list of bounding-boxes on a given image using pillow.

    Args:
        image: either a Numpy array or a PIL Image object
        bounding_boxes: List of BoundingBox objects
        font_path: String, path to the font file which should be used
        rgb_colors: List of colors, for matching a class-name to an index of the rgb-color list
        thickness: int, thickness of the box lines
        fill_background: Bool, whether to put the label on white background or not
        is_ground_truth: Bool, depending on ground truth or not, the label will be drawn
            in the upper left corner (no gt) or in the lower left corner (is gt)

    Returns:
        Numpy array, the given image decorated with annotations
    """

    white_color = (255, 255, 255, 15)

    if type(image) is np.ndarray:
        image = Image.fromarray(image)

    assert isinstance(image, Image.Image)

    font = ImageFont.truetype(
        font=font_path, size=np.floor(0.042 * image.size[1] + 0.1).astype("int32")
    )

    draw = ImageDraw.Draw(image)

    for bounding_box in bounding_boxes:
        label = "{} {:.2f}".format(bounding_box.class_name, bounding_box.score)
        label_size = draw.textsize(label, font)

        if is_ground_truth:
            text_origin = (
                float(bounding_box.box.xmin),
                float(bounding_box.box.ymax - label_size[1]),
            )
        else:
            if bounding_box.box.ymin - label_size[1] >= 0:
                text_origin = (
                    float(bounding_box.box.xmin),
                    float(bounding_box.box.ymin - label_size[1]),
                )
            else:
                text_origin = (
                    float(bounding_box.box.xmin),
                    float(bounding_box.box.ymin + 1),
                )

        color = rgb_colors[bounding_box.class_id]

        if fill_background:
            draw.rectangle(
                (
                    text_origin,
                    (text_origin[0] + label_size[0], text_origin[1] + label_size[1]),
                ),
                fill=white_color,
            )

            draw.text(text_origin, label, fill=(0, 0, 0, 15), font=font)
        else:
            draw.text(text_origin, label, fill=color, font=font)

        for draw_step in range(thickness):
            draw.rectangle(
                (
                    float(bounding_box.box.xmin + draw_step),
                    float(bounding_box.box.ymin + draw_step),
                    float(bounding_box.box.xmax - draw_step),
                    float(bounding_box.box.ymax - draw_step),
                ),
                outline=color,
            )

    del draw

    return np.asarray(image)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for building BaseAnnotation objects"""

from abc import ABC, abstractmethod
from typing import List

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.box import Box


class AnnotationBuilder(ABC):
    """
    Super class for defining the methods that are needed to build a single
    instance of an BaseAnnotation.
    """

    @staticmethod
    def _check_annotation(annotation: BaseAnnotation) -> None:
        wrong_bounding_boxes: List[Box] = []
        for bounding_box in annotation.bounding_boxes:
            if not annotation.is_valid_bounding_box(box=bounding_box.box):
                wrong_bounding_boxes.append(bounding_box.box)

        if wrong_bounding_boxes:
            raise ValueError(
                f"Annotation invalid \n"
                f"Wrong boxes: {wrong_bounding_boxes}\n"
                f"Annotation:\n{annotation}"
            )

        wrong_bounding_boxes = []
        for segmentation in annotation.segmentations:
            if segmentation.box is not None and not annotation.is_valid_bounding_box(
                box=segmentation.box
            ):
                wrong_bounding_boxes.append(segmentation.box)

        if wrong_bounding_boxes:
            raise ValueError(
                f"Annotation invalid \n"
                f"Wrong boxes: {wrong_bounding_boxes}\n"
                f"Annotation:\n{annotation}"
            )

    @abstractmethod
    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        """
        Builds a BaseAnnotation object utilizing the given parameters

        Args:
            image_path: String, points to an annotated image
            annotation_path: String, points to the respective annotation
            image_dir: String, points to the dir where the annotated image is stored
            annotation_dir: String, points to the dir where the respective annotation is stored
            replacement_string: String, part of the paths that is a placeholder

        Returns:
            A BaseAnnotation object
        """

        raise NotImplementedError()



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

""" Class for Bounding Box Annotation"""

from __future__ import annotations

import math
from typing import Any, Dict, List, Optional

from mlcvzoo_base.api.data.annotation_attributes import AnnotationAttributes
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification


class BoundingBox(AnnotationAttributes, Classification):
    """
    A class for defining the data object consumed by ObjectDetection models.
    It is mainly described by the box attribute, which covers an rectangular
    area of an image and is associated with a certain class
    """

    def __init__(
        self,
        box: Box,
        class_identifier: ClassIdentifier,
        score: float,
        difficult: bool,
        occluded: bool,
        content: str,
        model_class_identifier: Optional[ClassIdentifier] = None,
    ):
        Classification.__init__(
            self,
            class_identifier=class_identifier,
            model_class_identifier=model_class_identifier,
            score=score,
        )
        AnnotationAttributes.__init__(
            self, difficult=difficult, occluded=occluded, content=content
        )
        self.__box = box

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        if reduced:
            return {
                "box": self.box if raw_type else self.box.to_dict(),
                "class_id": self.class_id,
                "class_name": self.class_name,
                "model_class_id": self.model_class_identifier.class_id,
                "model_class_name": self.model_class_identifier.class_name,
                "score": self.score,
            }
        else:
            return {
                "box": self.box if raw_type else self.box.to_dict(),
                "class_identifier": self.class_identifier
                if raw_type
                else self.class_identifier.to_dict(),
                "model_class_identifier": self.model_class_identifier
                if raw_type
                else self.model_class_identifier.to_dict(),
                "score": self.score,
                "difficult": self.difficult,
                "occluded": self.occluded,
                "content": self.content,
            }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any], reduced: bool = False) -> BoundingBox:
        # fmt: off
        if reduced:
            return BoundingBox(**{
                "box": Box(**input_dict["box"]),
                "class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["class_id"],
                    "class_name": input_dict["class_name"],
                }),
                "model_class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["model_class_id"],
                    "class_name": input_dict["model_class_name"],
                }),
                "score": input_dict["score"],
                "difficult": False,
                "occluded": False,
                "content": "",
            })
        else:
            return BoundingBox(**{
                "box": Box(**input_dict["box"]),
                "class_identifier": ClassIdentifier(
                    **input_dict["class_identifier"]
                ),
                "model_class_identifier": ClassIdentifier(
                    **input_dict["model_class_identifier"]
                ),
                "score": input_dict["score"],
                "difficult": input_dict["difficult"],
                "occluded": input_dict["occluded"],
                "content": input_dict["content"],
            })
        # fmt: on

    def to_json(self) -> Any:
        return self.to_dict(raw_type=False)

    @property
    def box(self) -> Box:
        return self.__box

    def __eq__(self, other: BoundingBox):  # type: ignore
        # NOTE: Since floats may very for different systems, don't check the score for equality,
        #       but allow it to be in a reasonable range
        return (
            self.box == other.box
            and self.class_identifier.class_id == other.class_identifier.class_id
            and self.class_identifier.class_name == other.class_identifier.class_name
            and self.model_class_identifier.class_id == other.model_class_identifier.class_id
            and self.model_class_identifier.class_name == other.model_class_identifier.class_name
            and self.occluded == other.occluded
            and self.difficult == other.difficult
            and self.content == other.content
            and math.isclose(a=self.score, b=other.score, abs_tol=0.005)
        )

    def __repr__(self):  # type: ignore
        return (
            f"BoundingBox: "
            f"class-id={self.class_id}, "
            f"class-name={self.class_name}: "
            f"model-class-id={self.model_class_identifier.class_id}, "
            f"model-class-name={self.model_class_identifier.class_name}: "
            f"Box={self.box}, "
            f"score={self.score}, "
            f"difficult={self.difficult}, "
            f"occluded={self.occluded}, "
            f"content='{self.content}'"
        )

    def to_list(self) -> List[int]:
        """
        Transforms the BoundingBox object to a list of its coordinates.

        Returns:
            A 1x4 list of the objects coordinates [xmin, ymin, xmax, ymax]
        """
        return [self.box.xmin, self.box.ymin, self.box.xmax, self.box.ymax]

    def copy_bounding_box(self, class_identifier: ClassIdentifier) -> BoundingBox:
        return BoundingBox(
            box=self.box,
            class_identifier=class_identifier,
            score=self.score,
            difficult=self.difficult,
            occluded=self.occluded,
            content=self.content,
            model_class_identifier=self.model_class_identifier,
        )



# --------------------------------------------------------
# Copyright (c) 2015-2016 Adrian Rosebrock
# Licensed under The MIT License [see LICENSE.txt for details]
# author:    Adrian Rosebrock
# website:   http://www.pyimagesearch.com
# --------------------------------------------------------
"""
Module for implementations taken from third party package 'imutils'
https://github.com/PyImageSearch/imutils
"""
import numpy as np
from scipy.spatial import distance as dist


def order_points(polygon: np.ndarray, sort_by_euclidean: bool = True) -> np.ndarray:  # type: ignore[type-arg]
    """
    Orders points of the given polygon in the order:
    [top left, top right, bottom right, bottom left]

    Base of the code taken from:
    https://github.com/PyImageSearch/imutils/blob/
    c12f15391fcc945d0d644b85194b8c044a392e0a/imutils/perspective.py#L9-L34

    Args:
        polygon: a Numpy array, containing polygon defining edge points
        sort_by_euclidean: Whether or not to sort the rightMost points by using the
                           euclidean distance between points

    Returns:
        A Numpy array, a permuted version of the input array entries
    """

    # sort the points based on their x-coordinates
    sorted_polygon = polygon[np.argsort(polygon[:, 0]), :]

    # grab the left-most and right-most points from the sorted
    # x-coordinate points
    leftMost = sorted_polygon[:2, :]
    rightMost = sorted_polygon[2:, :]

    # now, sort the left-most coordinates according to their
    # y-coordinates so we can grab the top-left and bottom-left
    # points, respectively
    leftMost = leftMost[np.argsort(leftMost[:, 1]), :]
    (tl, bl) = leftMost

    # NOTE: the sorting on the basis of the euclidean distance does not work
    #       in all cases. That is the reason why we duplicated this method to
    #       the MLCVZoo.
    # TODO: add a pull request in https://github.com/PyImageSearch/imutils
    if sort_by_euclidean:
        # now that we have the top-left coordinate, use it as an
        # anchor to calculate the Euclidean distance between the
        # top-left and right-most points; by the Pythagorean
        # theorem, the point with the largest distance will be
        # our bottom-right point
        D = dist.cdist(tl[np.newaxis], rightMost, "euclidean")[0]
        (br, tr) = rightMost[np.argsort(D)[::-1], :]
    else:
        # now, sort the right-most coordinates according to their
        # y-coordinates so we can grab the top-left and bottom-left
        # points, respectively
        rightMost = rightMost[np.argsort(rightMost[:, 1]), :]
        (tr, br) = rightMost

    # return the coordinates in top-left, top-right,
    # bottom-right, and bottom-left order
    return np.array([tl, tr, br, bl], dtype="float32")



# Copyright 2022 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing Label Studio formatted annotations"""

import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError, ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerLabelStudioInputDataConfig,
)
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.utils import get_file_list

logger = logging.getLogger(__name__)


class LabelStudioAnnotationParser(AnnotationParser):
    """
    Parser for the Label Studio json format.

    REMARK:
    For now only Bounding Boxes are supported
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        label_studio_input_data: List[AnnotationHandlerLabelStudioInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.label_studio_input_data = label_studio_input_data

    @staticmethod
    def __get_label_studio_annotation_dict(
        label_studio_annotation_path: str,
    ) -> Optional[Dict[str, Any]]:
        with open(label_studio_annotation_path, "r") as label_studio_file:
            try:
                return cast(Dict[str, Any], json.load(label_studio_file))
            except json.decoder.JSONDecodeError as json_error:
                logger.exception(
                    msg="Could not parse annotation file '%s'. Annotation will be skipped."
                    % label_studio_annotation_path,
                    exc_info=json_error,
                )

        return None

    def parse(self) -> List[BaseAnnotation]:

        annotations: List[BaseAnnotation] = []

        for dataset_count, input_data in enumerate(self.label_studio_input_data):
            input_image_types = input_data.image_format.split("|")

            input_image_paths: List[str] = []
            for input_image_type in input_image_types:
                input_image_paths.extend(
                    get_file_list(
                        input_dir=input_data.input_image_dir,
                        search_subfolders=True,
                        file_extension=input_image_type,
                    )
                )

            image_file_dict = {os.path.basename(p): p for p in input_image_paths}
            label_studio_annotation_paths = [
                str(p) for p in Path(input_data.input_annotation_dir).glob("**/*")
            ]

            for label_studio_annotation_path in label_studio_annotation_paths:
                label_studio_dict = LabelStudioAnnotationParser.__get_label_studio_annotation_dict(
                    label_studio_annotation_path=label_studio_annotation_path
                )
                if label_studio_dict is None:
                    continue

                try:
                    label_studio_image_name = os.path.basename(
                        label_studio_dict["task"]["data"]["image"]
                    )

                    if label_studio_image_name in image_file_dict:
                        image_path = image_file_dict[label_studio_image_name]
                    else:
                        logger.warning(
                            "Label Studio annotation contains non-existing image: '%s', "
                            "Annotation will be skipped." % label_studio_image_name
                        )
                        continue

                    image_shape: Optional[Tuple[int, int]] = None
                    bounding_boxes: List[BoundingBox] = []
                    for result in label_studio_dict["result"]:
                        if image_shape is None:
                            image_shape = (result["original_height"], result["original_width"])

                        if result["type"] == "rectanglelabels" and image_shape is not None:
                            try:
                                bounding_box: Optional[BoundingBox] = self.__parse_bounding_box(
                                    label_studio_value=result["value"], image_shape=image_shape
                                )
                                if bounding_box is not None:
                                    bounding_boxes.append(bounding_box)

                            except (ValueError, ForbiddenClassError) as error:
                                logger.warning("%s, annotation will be skipped" % str(error))
                                continue

                    if image_shape is not None:
                        annotations.append(
                            BaseAnnotation(
                                image_path=image_path,
                                annotation_path=label_studio_annotation_path,
                                image_shape=image_shape,
                                classifications=[],
                                bounding_boxes=bounding_boxes,
                                segmentations=[],
                                image_dir=os.path.dirname(image_path),
                                annotation_dir=os.path.dirname(label_studio_annotation_path),
                                # TODO: How to set this?
                                replacement_string="",
                            )
                        )
                except KeyError as index_error:
                    logger.exception(
                        msg="Could not parse annotation file '%s'. Annotation will be skipped."
                        % label_studio_annotation_path,
                        exc_info=index_error,
                    )

        return annotations

    def __parse_bounding_box(
        self, label_studio_value: Dict[str, Any], image_shape: Tuple[int, int]
    ) -> Optional[BoundingBox]:

        label_studio_class_name: str = label_studio_value["rectanglelabels"][0]

        try:
            class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                class_name=label_studio_class_name
            )
            class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                class_name=label_studio_class_name
            )

            return BoundingBox(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=1.0,
                difficult=False,
                occluded=False,
                content="",
                box=Box.init_format_based(
                    box_list=(
                        int(image_shape[1] * label_studio_value["x"] / 100),
                        int(image_shape[0] * label_studio_value["y"] / 100),
                        int(image_shape[1] * label_studio_value["width"] / 100),
                        int(image_shape[0] * label_studio_value["height"] / 100),
                    ),
                    box_format=ObjectDetectionBBoxFormats.XYWH,
                ),
            )

        except ClassMappingNotFoundError:
            logger.warning(
                "Could not find a valid class-mapping for class-name '%s'. "
                "BoundingBox will be skipped",
                label_studio_class_name,
            )
            return None



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Data class that holds attributes of Classification objects"""

from __future__ import annotations

import math
from typing import Any, Dict, Optional

from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.interfaces import Perception


class Classification(Perception):
    """
    Class which is used to state which object type is described by an given image.
    """

    # ClassIdentifier attribute that represents the primary class information
    # of a Classification object
    _class_identifier: ClassIdentifier

    # ClassIdentifier attribute that is used to store class information that has
    # been produced by a model. In most cases it is equal to the 'class_identifier',
    # but in cases where the class information of a models prediction has been
    # mapped / post processed, it can be used to access the real output of the model.
    _model_class_identifier: ClassIdentifier

    # score which expresses the likelihood that the class-id / class-name is correct
    __score: float

    def __init__(
        self,
        class_identifier: ClassIdentifier,
        score: float,
        model_class_identifier: Optional[ClassIdentifier] = None,
    ):
        self.__score = score
        self.__class_identifier = class_identifier

        if model_class_identifier is None:
            self.__model_class_identifier = class_identifier
        else:
            self.__model_class_identifier = model_class_identifier

    def __eq__(self, other: Classification) -> bool:  # type: ignore
        return (
            # 4 decimals should be plenty for accuracy
            math.isclose(a=self.score, b=other.score, abs_tol=0.0005)
            and self.__class_identifier.class_id == other.__class_identifier.class_id
            and self.__class_identifier.class_name == other.__class_identifier.class_name
            and self.__model_class_identifier.class_id == other.__model_class_identifier.class_id
            and self.__model_class_identifier.class_name
            == other.__model_class_identifier.class_name
        )

    def __repr__(self):  # type: ignore
        return (
            f"Classification("
            f"class-id={self.class_id}, "
            f"class-name={self.class_name}: "
            f"model-class-id={self.model_class_identifier.class_id}, "
            f"model-class-name={self.model_class_identifier.class_name}: "
            f"score={self.score})"
        )

    @property
    def class_identifier(self) -> ClassIdentifier:
        return self.__class_identifier

    @property
    def model_class_identifier(self) -> ClassIdentifier:
        return self.__model_class_identifier

    @property
    def class_id(self) -> int:
        return self.__class_identifier.class_id

    @property
    def class_name(self) -> str:
        return self.__class_identifier.class_name

    @property
    def score(self) -> float:
        return self.__score

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        if reduced:
            return {
                "class_id": self.class_id,
                "class_name": self.class_name,
                "model_class_id": self.model_class_identifier.class_id,
                "model_class_name": self.model_class_identifier.class_name,
                "score": self.score,
            }
        else:
            return {
                "class_identifier": self.class_identifier
                if raw_type
                else self.class_identifier.to_dict(),
                "model_class_identifier": self.model_class_identifier
                if raw_type
                else self.model_class_identifier.to_dict(),
                "score": self.score,
            }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any], reduced: bool = False) -> Classification:
        # fmt: off
        if reduced:
            return Classification(**{
                "class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["class_id"],
                    "class_name": input_dict["class_name"],
                }),
                "model_class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["model_class_id"],
                    "class_name": input_dict["model_class_name"],
                }),
                "score": input_dict["score"],
            })
        else:
            return Classification(**{
                "class_identifier": ClassIdentifier(
                    **input_dict["class_identifier"]
                ),
                "model_class_identifier": ClassIdentifier(
                    **input_dict["model_class_identifier"]
                ),
                "score": input_dict["score"],
            })
        # fmt: on

    def copy_classification(self, class_identifier: ClassIdentifier) -> Classification:
        return Classification(
            class_identifier=class_identifier,
            score=self.score,
            model_class_identifier=self.model_class_identifier,
        )



# Copyright 2022 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for defining a component that enables to create registries for generic class types
"""

import importlib
import logging
from typing import Dict, Generic, TypeVar

ClassType = TypeVar("ClassType")

logger = logging.getLogger(__name__)


class MLCVZooRegistry(Generic[ClassType]):
    """
    Basis class that enables to create registries for generic class types
    """

    def __init__(self) -> None:
        self._registry: Dict[str, ClassType] = {}

    def register_module(
        self, module_type_name: str, module_constructor: ClassType, force: bool = False
    ) -> None:
        """
        Directly register an module constructor

        Args:
            module_type_name: Name of the model to register
            module_constructor: The reference to the model constructor
            force: Overwrite an existing entry

        Returns:
            None
        """

        if not force and module_type_name in self._registry:
            raise KeyError(
                f"{module_type_name} is already registered model registry" f"in {self._registry}"
            )

        self._registry[module_type_name] = module_constructor

    def register_external_module(
        self, module_type_name: str, module_constructor: str, package_name: str
    ) -> None:
        """
        Register an external model

        Args:
            module_type_name: Name of the model to register
            module_constructor: Name of the constructor of the model to register as string
            package_name: The full package to import to call the constructor

        Returns:
            None
        """
        try:
            # pylint: disable=c0415
            module = importlib.import_module(package_name)
            self.register_module(
                module_type_name=module_type_name,
                module_constructor=module.__dict__[module_constructor],
            )
        except ImportError as import_error:
            logger.info(
                "Optional module '%s' (%s.%s) not available: %s"
                % (module_type_name, package_name, module_constructor, import_error)
            )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

""" Base class for image annotations """

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.ocr_perception import OCRPerception
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.data_preparation.structs import CSVOutputStringFormats

logger = logging.getLogger(__name__)


# TODO: Ensure annotations and contained coordinates are valid at any point in time
@dataclass
class BaseAnnotation:
    """
    An image annotation.
    """

    # Absolute file path to the image of the annotation
    image_path: str
    # Absolute file path for the annotation
    annotation_path: str

    # shape of the image in the format (height, width)
    image_shape: Tuple[int, int]
    # List of classifications for this image. Can be used for training classification algorithms,
    # or to filter out images by their content.
    classifications: List[Classification] = field(default_factory=lambda: [])
    # List of bounding_boxes
    bounding_boxes: List[BoundingBox] = field(default_factory=lambda: [])
    # List of segmentations
    segmentations: List[Segmentation] = field(default_factory=lambda: [])
    # List of ocr perceptions (Text)
    ocr_perception: Optional[OCRPerception] = None

    image_dir: str = ""
    # TODO: rename to annotation_path_dir?
    #       When pascal-voc is used, the annotation_path_dir is a
    #       directory, when other formats are used,
    #       the annotation_path_dir is a path
    annotation_dir: str = ""
    replacement_string: str = ""

    def __repr__(self):  # type: ignore
        annotation_string = f"image_shape: {self.image_shape}\n"
        annotation_string += f"image-path: {self.image_path}\n"
        annotation_string += f"image-dir: {self.image_dir}\n"
        annotation_string += f"annotation-path: {self.annotation_path}\n"
        annotation_string += f"annotation-dir: {self.annotation_dir}\n"

        for classification in self.classifications:
            annotation_string += f"{classification}\n"

        for bounding_box in self.bounding_boxes:
            annotation_string += f"{bounding_box}\n"

        for segmentation in self.segmentations:
            annotation_string += f"{segmentation}\n"

        return annotation_string

    def __eq__(self, other):  # type: ignore
        if len(self.classifications) != len(other.classifications):
            return False
        if len(self.bounding_boxes) != len(other.bounding_boxes):
            return False
        if len(self.segmentations) != len(other.segmentations):
            return False

        for self_classification, other_classification in zip(
            self.classifications, other.classifications
        ):
            if self_classification != other_classification:
                return False

        for self_bounding_box, other_bounding_box in zip(
            self.bounding_boxes, other.bounding_boxes
        ):
            if self_bounding_box != other_bounding_box:
                return False

        for self_segmentation, other_segmentation in zip(self.segmentations, other.segmentations):
            if self_segmentation != other_segmentation:
                return False

        return True

    def to_dict(self) -> Dict[Any, Any]:
        return self.__dict__

    def get_height(self) -> int:
        """

        Returns:
            Height in pixels of the image for which the annotation is.

        """
        return int(self.image_shape[0])

    def get_width(self) -> int:
        """

        Returns:
            Width in pixels of the image for which the annotation is.

        """
        return int(self.image_shape[1])

    def get_bounding_boxes(self, include_segmentations: bool = True) -> List[BoundingBox]:
        """

        Args:
            include_segmentations (bool): Whether to return segmentations in the image

        Returns:
            list of BoundingBox objects. The bounding boxes (and segmentations) present
            in the image.

        """

        bounding_boxes: List[BoundingBox] = []

        bounding_boxes.extend([b for b in self.bounding_boxes if b.box is not None])

        if include_segmentations:
            bounding_boxes.extend(
                [s.to_bounding_box(self.image_shape) for s in self.segmentations]
            )

        return bounding_boxes

    def to_csv_entry(
        self,
        use_difficult: bool,
        use_occluded: bool,
        include_surrounding_bboxes: bool = True,
        output_string_format: str = CSVOutputStringFormats.BASE,
    ) -> Optional[str]:
        """
        Transforms the BaseAnnotation object to CSV format.

        Args:
            use_difficult (bool): Whether or not to consider the BaseAnnotation object if marked
            with 'difficult' flag
            use_occluded(bool): Whether or not to consider the BaseAnnotation object marked
            with 'occluded' flag
            include_surrounding_bboxes (bool): Whether or not to consider the BaseAnnotation
            object if it is a segmentation.
            output_string_format (str): One of CSVOutputStringFormats. Defines the format
            type of the resulting CSV string.

        Returns:
            String in CSV Format, Optional. Depending on the input parameters and format type
            the method either returns a CSV sting or None.

        """

        if output_string_format == CSVOutputStringFormats.BASE:
            csv_entry = self.__to_base_entry(
                use_difficult=use_difficult,
                use_occluded=use_occluded,
                include_segmentations=include_surrounding_bboxes,
            )
        elif output_string_format == CSVOutputStringFormats.YOLO:
            csv_entry = self.__to_yolo_entry(
                use_difficult=use_difficult,
                use_occluded=use_occluded,
                include_segmentations=include_surrounding_bboxes,
            )
        else:
            logger.warning(
                "Could not find a valid output_string_format. Given format: %s",
                output_string_format,
            )
            csv_entry = ""

        return csv_entry

    def __to_base_entry(
        self, use_difficult: bool, use_occluded: bool, include_segmentations: bool
    ) -> Optional[str]:
        csv_line = ""

        for bounding_box in self.get_bounding_boxes(include_segmentations=include_segmentations):
            if not use_occluded and bounding_box.occluded:
                logger.debug("Skip occluded bounding-box: %r", bounding_box)
                continue

            if not use_difficult and bounding_box.difficult:
                logger.debug("Skip difficult bounding-box: %r", bounding_box)
                continue

            if bounding_box.box is not None:
                is_valid_box = self.is_valid_bounding_box(box=bounding_box.box)
                if is_valid_box:
                    csv_line += "{},{},{},{},{} ".format(
                        bounding_box.box.xmin,
                        bounding_box.box.ymin,
                        bounding_box.box.xmax,
                        bounding_box.box.ymax,
                        bounding_box.class_name,
                    )
                else:
                    logger.warning(
                        "Skip bounding-box because it doesn't fulfill "
                        "the requirements check by is_valid_bounding_box: \n"
                        "   bounding-box: %s",
                        bounding_box,
                    )

        if csv_line != "":
            csv_entry: Optional[str] = "{} {} {} {}\n".format(
                self.image_path, self.image_shape[0], self.image_shape[1], csv_line
            )
        else:
            csv_entry = None

        return csv_entry

    def __to_yolo_entry(
        self, use_difficult: bool, use_occluded: bool, include_segmentations: bool
    ) -> Optional[str]:
        csv_line = ""

        for bounding_box in self.get_bounding_boxes(include_segmentations=include_segmentations):
            if not use_occluded and bounding_box.occluded:
                logger.debug("Skip occluded bounding-box: %s", bounding_box)
                continue

            if not use_difficult and bounding_box.difficult:
                logger.debug("Skip difficult bounding-box: %s", bounding_box)
                continue

            if bounding_box.box is not None:
                is_valid_box = self.is_valid_bounding_box(box=bounding_box.box)
                if is_valid_box:
                    csv_line += "{},{},{},{},{} ".format(
                        bounding_box.box.xmin,
                        bounding_box.box.ymin,
                        bounding_box.box.xmax,
                        bounding_box.box.ymax,
                        bounding_box.class_id,
                    )
                else:
                    logger.warning(
                        "Skip bounding-box because it doesn't fulfill "
                        "the requirements check by is_valid_bounding_box: \n"
                        "   bounding-box: %s",
                        bounding_box,
                    )

        if csv_line != "":
            csv_entry: Optional[str] = "{} {}\n".format(self.image_path, csv_line)
        else:
            csv_entry = None

        return csv_entry

    def is_valid_bounding_box(self, box: Box) -> bool:
        """

        Args:
            box: a Box object

        Returns:
            True if the coordinates of the Annotation are valid, else False.

        """

        bbox_has_valid_coordinates: bool = (
            (0 <= box.xmin <= self.get_width())
            and (0 <= box.xmax <= self.get_width())
            and (0 <= box.ymin <= self.get_height())
            and (0 <= box.ymax <= self.get_height())
            and (box.xmin < box.xmax)
            and (box.ymin < box.ymax)
        )

        return bbox_has_valid_coordinates

    def copy_annotation(
        self,
        classifications: Optional[List[Classification]] = None,
        bounding_boxes: Optional[List[BoundingBox]] = None,
        segmentations: Optional[List[Segmentation]] = None,
    ) -> BaseAnnotation:
        return BaseAnnotation(
            image_path=self.image_path,
            annotation_path=self.image_path,
            image_shape=self.image_shape,
            classifications=classifications
            if classifications is not None
            else self.classifications,
            bounding_boxes=bounding_boxes if bounding_boxes is not None else self.bounding_boxes,
            segmentations=segmentations if segmentations is not None else self.segmentations,
            image_dir=self.image_dir,
            annotation_dir=self.annotation_dir,
            replacement_string=self.replacement_string,
        )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

import copy
import logging
from typing import Dict, List, Optional, Tuple

import numpy as np
from tqdm import tqdm

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.evaluation.object_detection.data_classes import (
    CONFUSION_MATRIX_DICT_TYPE,
    CONFUSION_MATRIX_TYPE,
    DEFAULT_FLOAT_VALUE,
    DEFAULT_INT_VALUE,
    EVALUATION_LIST_TYPE,
    METRIC_DICT_TYPE,
    METRIC_IMAGE_INFO_TYPE,
    MetricImageInfo,
    ODEvaluationComputingData,
    ODMetrics,
    ODModelEvaluationMetrics,
)
from mlcvzoo_base.evaluation.object_detection.structs import BBoxSizeTypes
from mlcvzoo_base.evaluation.object_detection.utils import (
    compute_max_bounding_box,
    generate_metric_table,
    get_bbox_size_type,
)
from mlcvzoo_base.third_party.py_faster_rcnn.voc_ap import voc_ap

logger = logging.getLogger(__name__)


class MetricsComputation:
    """
    Main class for handling the evaluation of object detection models.
    """

    iou_thresholds_ap_50_95: List[float] = [
        0.5,
        0.55,
        0.6,
        0.65,
        0.7,
        0.75,
        0.8,
        0.85,
        0.9,
        0.95,
    ]

    def __init__(
        self,
        model_specifier: str,
        classes_id_dict: Optional[Dict[int, str]],
        iou_thresholds: List[float],
        gt_annotations: List[BaseAnnotation],
        predicted_bounding_boxes_list: List[List[BoundingBox]],
        mapper: Optional[AnnotationClassMapper] = None,
    ):
        self.model_specifier = model_specifier
        self.iou_thresholds: List[float] = iou_thresholds
        self.dataset_length: int = len(gt_annotations)

        self.class_identifier_list: List[ClassIdentifier] = []
        self._classes_id_dict = classes_id_dict
        if self._classes_id_dict is not None:
            logger.warning(
                "DEPRECATED: The parameter classes_id_dict is deprecated and will"
                "be removed in future versions, please use mapper instead"
            )
            for class_id, class_name in self._classes_id_dict.items():
                self.class_identifier_list.append(
                    ClassIdentifier(class_id=class_id, class_name=class_name)
                )
        else:
            if mapper is not None:
                self.mapper = mapper
                self.class_identifier_list = self.mapper.create_class_identifier_list()
            else:
                raise ValueError(
                    "Either the parameter 'mapper' or the parameter 'classes_id_dict'(DEPRECATED) "
                    "has to be provided "
                )

        # Remove duplicates
        self.class_identifier_list = list(dict.fromkeys(self.class_identifier_list))

        self._all_predicted_annotations_dict: Dict[int, Dict[str, List[BaseAnnotation]]] = {
            i: {} for i in range(self.dataset_length)
        }
        self._all_gt_annotations_dict: Dict[int, Dict[str, List[BaseAnnotation]]] = {
            i: {} for i in range(self.dataset_length)
        }

        self.model_metrics: ODModelEvaluationMetrics = ODModelEvaluationMetrics(
            model_specifier=self.model_specifier
        )
        self.computing_data: ODEvaluationComputingData = ODEvaluationComputingData()

        process_bar = tqdm(
            zip(gt_annotations, predicted_bounding_boxes_list),
            desc=f"Compute metrics",
        )

        # TODO: Convert to batch-processing
        for index, (gt_annotation, predicted_bounding_boxes) in enumerate(process_bar):
            _ = self.__update_from_prediction(
                index=index,
                gt_annotation=gt_annotation,
                predicted_bounding_boxes=predicted_bounding_boxes,
            )

    @property
    def classes_id_dict(self) -> Optional[Dict[int, str]]:
        logger.warning(
            "DEPRECATED: The attribute 'classes_id_dict' is deprecated"
            "and will be removed in future versions"
        )
        return self._classes_id_dict

    @property
    def all_predicted_annotations(self) -> EVALUATION_LIST_TYPE:
        logger.warning(
            "DEPRECATED: The attribute 'all_predicted_annotations' is deprecated"
            "and will be removed in future versions"
        )
        return [[]]

    @property
    def all_gt_annotations(self) -> EVALUATION_LIST_TYPE:
        logger.warning(
            "DEPRECATED: The attribute 'all_gt_annotations' is deprecated"
            "and will be removed in future versions"
        )
        return [[]]

    def get_metrics_dict(self) -> METRIC_DICT_TYPE:
        return self.model_metrics.metrics_dict

    def get_metrics_image_info_dict(
        self,
    ) -> METRIC_IMAGE_INFO_TYPE:
        return self.model_metrics.metrics_image_info_dict

    @staticmethod
    def get_overall_ap(metrics_dict: METRIC_DICT_TYPE, iou_threshold: float) -> float:
        """
        Calculate AP metric over every class specific AP metric for bounding boxes of all sizes.

        Args:
            metrics_dict: The dictionary that stores the relevant metric data, which is the
                          basis for the calculation
            iou_threshold: The iou threshold for which the AP metric should be computed

        Returns:
            The computed overall AP metric
        """

        if iou_threshold in metrics_dict:
            class_metrics = metrics_dict[iou_threshold][BBoxSizeTypes.BBOX_ALL]
        else:
            raise ValueError(
                "Can not compute overall AP for iou-threshold=%s, "
                "no data is given in the metrics-dict.",
                iou_threshold,
            )

        ap_list: List[float] = [class_metric.AP for class_metric in class_metrics.values()]
        return sum(ap_list) / len(ap_list)

    @staticmethod
    def compute_average_ap(model_metrics: ODModelEvaluationMetrics) -> float:
        """
        Compute the average of the AP metric for every
        overall AP (average over all classes) of all iou-thresholds
        in the metrics_dict.

        Args:
            model_metrics: The model-metrics for which to compute the average ap

        Returns:
            The computed average AP metric
        """
        ap_list: List[float] = [
            MetricsComputation.get_overall_ap(
                metrics_dict=model_metrics.metrics_dict,
                iou_threshold=iou_threshold,
            )
            for iou_threshold in model_metrics.metrics_dict.keys()
        ]
        return sum(ap_list) / len(ap_list)

    @staticmethod
    def get_ap_50_95(model_metrics: ODModelEvaluationMetrics) -> float:
        """
        Compute the COCO mAP metric which is defined as the AP
        metric for the iou-thresholds = [0.5, 0.55, ..., 0.95].

        Args:
            model_metrics: The model-metrics for which to compute the COCO mAP

        Returns:
            The computed COCO mAP
        """
        return sum(
            [
                MetricsComputation.get_overall_ap(
                    metrics_dict=model_metrics.metrics_dict, iou_threshold=iou
                )
                for iou in MetricsComputation.iou_thresholds_ap_50_95
            ]
        ) / len(MetricsComputation.iou_thresholds_ap_50_95)

    @staticmethod
    def get_ap_50(model_metrics: ODModelEvaluationMetrics) -> float:
        """
        Compute the AP50 metric which is defined as the AP
        metric for the iou-threshold=0.5

        Args:
            model_metrics: The model-metrics for which to compute the COCO mAP

        Returns:
            The computed AP50
        """

        return MetricsComputation.get_overall_ap(
            metrics_dict=model_metrics.metrics_dict, iou_threshold=0.5
        )

    def __reset_main_dictionaries(
        self,
    ) -> None:
        for class_identifier in self.class_identifier_list:
            self.model_metrics.metrics_image_info_dict[str(class_identifier)] = dict()

        for bbox_size_type in BBoxSizeTypes.get_values_as_list(class_type=BBoxSizeTypes):
            self.computing_data.gt_counter_dict[bbox_size_type] = dict()

            for class_identifier in self.class_identifier_list:
                self.computing_data.gt_counter_dict[bbox_size_type][str(class_identifier)] = 0

        for iou_thresh in self.iou_thresholds:
            self.model_metrics.metrics_dict[iou_thresh] = dict()
            self.computing_data.valid_precisions[iou_thresh] = dict()

            for bbox_size_type in BBoxSizeTypes.get_values_as_list(class_type=BBoxSizeTypes):
                self.model_metrics.metrics_dict[iou_thresh][bbox_size_type] = dict()
                self.computing_data.valid_precisions[iou_thresh][bbox_size_type] = list()

                # TODO: Make None to state that now data has been computed?
                for class_identifier in self.class_identifier_list:
                    self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][
                        str(class_identifier)
                    ] = ODMetrics()

    def __reset_computing_dictionaries(self) -> None:
        for iou_thresh in self.iou_thresholds:
            self.computing_data.false_positives_dict[iou_thresh] = dict()
            self.computing_data.true_positives_dict[iou_thresh] = dict()
            self.computing_data.detected_annotations[iou_thresh] = list()
            self.computing_data.scores[iou_thresh] = dict()

            for bbox_size in BBoxSizeTypes.get_values_as_list(class_type=BBoxSizeTypes):
                self.computing_data.false_positives_dict[iou_thresh][bbox_size] = np.zeros((0,))
                self.computing_data.true_positives_dict[iou_thresh][bbox_size] = np.zeros((0,))

                self.computing_data.scores[iou_thresh][bbox_size] = np.zeros((0,))

    def compute_metrics(
        self,
    ) -> ODModelEvaluationMetrics:
        """
        Compute MAP Metrics

        return: the metrics dictionary in the form of:
                1st key: iou-threshold
                2nd key: type of the size of the bounding-box
                3rd key: class-name
                Value: The computed metrics

                Dict[IOU_THRESHOLD][BBoxSizeTypes.BBOX_SIZE_TYPE][CLASS_NAME] = ODMetrics
        """

        self.__reset_main_dictionaries()

        # process get_bounding_boxes and annotations
        for class_identifier in self.class_identifier_list:
            self.__reset_computing_dictionaries()

            self.__fill_computing_data(
                class_id=class_identifier.class_id,
                class_name=class_identifier.class_name,
                class_identifier=class_identifier,
            )

            self.__compute_metrics(class_identifier=class_identifier)

        for iou_threshold in self.iou_thresholds:
            metric_table = generate_metric_table(
                metrics_dict=self.model_metrics.metrics_dict,
                iou_threshold=iou_threshold,
            )
            logger.debug(
                "Metrics for IOU-Threshold '%s': \n%s" % (iou_threshold, metric_table.table)
            )

        return ODModelEvaluationMetrics(
            model_specifier=self.model_specifier,
            metrics_dict=self.get_metrics_dict(),
            metrics_image_info_dict=self.get_metrics_image_info_dict(),
        )

    def __update_from_prediction(
        self,
        index: int,
        gt_annotation: BaseAnnotation,
        predicted_bounding_boxes: List[BoundingBox],
    ) -> BaseAnnotation:
        """
        Update the main computing attributes given the information from the ground truth
        annotation and predicted bounding boxes at the dataset index.

        Args:
            index: The index of the dataset where to update the data
            gt_annotation: The ground truth information for the given index (data item)
            predicted_bounding_boxes: The predicted bounding boxes for this index (data item)

        Returns:
            The annotation object created on the basis of the ground truth data and the predicted
            bounding boxes
        """

        logger.debug(
            "Get metrics for annotation: \n"
            "  - image-path:       %s\n"
            "  - image-annotation: %s",
            gt_annotation.image_path,
            gt_annotation.annotation_path,
        )

        self._all_gt_annotations_dict = MetricsComputation.__update_annotation_data_dict(
            class_identifier_list=self.class_identifier_list,
            all_annotations_dict=self._all_gt_annotations_dict,
            index=index,
            new_annotation=gt_annotation,
        )

        predicted_annotation = BaseAnnotation(
            image_path=gt_annotation.image_path,
            annotation_path=gt_annotation.annotation_path,
            image_shape=gt_annotation.image_shape,
            classifications=[],
            bounding_boxes=predicted_bounding_boxes,
            segmentations=[],
            image_dir=gt_annotation.image_dir,
            annotation_dir=gt_annotation.annotation_dir,
            replacement_string=gt_annotation.replacement_string,
        )

        self._all_predicted_annotations_dict = MetricsComputation.__update_annotation_data_dict(
            class_identifier_list=self.class_identifier_list,
            all_annotations_dict=self._all_predicted_annotations_dict,
            index=index,
            new_annotation=predicted_annotation,
        )

        return predicted_annotation

    @staticmethod
    def __update_annotation_data_dict(
        class_identifier_list: List[ClassIdentifier],
        all_annotations_dict: Dict[int, Dict[str, List[BaseAnnotation]]],
        index: int,
        new_annotation: BaseAnnotation,
    ) -> Dict[int, Dict[str, List[BaseAnnotation]]]:
        """
        Update annotation information in the given 'all_annotations' list at the given index
        based on the annotation data given by 'new_annotation'.

        Args:
            class_identifier_list: List of class identifier for which to update the annotation
                                   data
            all_annotations_dict: The dictionary to update
            index: int, index of entry (image) which annotations should be updated
            new_annotation: BaseAnnotation object, holding the new annotation information

        Returns:
            The updated all_annotations_dict

        """

        if len(new_annotation.get_bounding_boxes(include_segmentations=True)) == 0:
            # No update necessary
            return all_annotations_dict

        # Gather (BoundingBox / Annotation) data for each class-id.
        # This is needed to be able to compute detailed metrics
        # for each class-id.
        for class_identifier in class_identifier_list:
            # Have to find all bounding-boxes for a specific class-id.
            # In order to have access to the information about the image-path,
            # we need to store an annotation object, rather than a single bounding-box.
            # Despite the fact that for the metric calculation itself the bounding-box
            # information is enough.
            # For purposes like: draw/log all false-positive bounding-boxes for specific
            # images we need to have the information about the image-path
            #   => therefore use BaseAnnotation and not BoundingBox alone
            all_annotations_dict[index][str(class_identifier)] = []
            for b in new_annotation.get_bounding_boxes(include_segmentations=True):
                if b.class_id == class_identifier.class_id:
                    metric_annotation = BaseAnnotation(
                        image_path=new_annotation.image_path,
                        annotation_path=new_annotation.annotation_path,
                        image_shape=new_annotation.image_shape,
                        classifications=[],
                        bounding_boxes=[
                            BoundingBox(
                                class_identifier=b.class_identifier,
                                model_class_identifier=b.model_class_identifier,
                                score=b.score,
                                box=b.box,
                                difficult=False,
                                occluded=False,
                                content="",
                            )
                        ],
                        segmentations=[],
                        image_dir=new_annotation.image_dir,
                        annotation_dir=new_annotation.annotation_dir,
                        replacement_string=new_annotation.replacement_string,
                    )

                    all_annotations_dict[index][str(class_identifier)].append(metric_annotation)

        return all_annotations_dict

    def __update_tp_fp_data(
        self,
        is_true_positive: bool,
        iou_thresh: float,
        bbox_size_type: str,
    ) -> None:
        if is_true_positive:
            tp_value = 1
            fp_value = 0
        else:
            tp_value = 0
            fp_value = 1

        # Update dictionary for overall mAP computation
        self.computing_data.true_positives_dict[iou_thresh][BBoxSizeTypes.BBOX_ALL] = np.append(
            self.computing_data.true_positives_dict[iou_thresh][BBoxSizeTypes.BBOX_ALL],
            tp_value,
        )
        self.computing_data.false_positives_dict[iou_thresh][BBoxSizeTypes.BBOX_ALL] = np.append(
            self.computing_data.false_positives_dict[iou_thresh][BBoxSizeTypes.BBOX_ALL],
            fp_value,
        )

        # Update dictionary for size specific mAP computation
        self.computing_data.true_positives_dict[iou_thresh][bbox_size_type] = np.append(
            self.computing_data.true_positives_dict[iou_thresh][bbox_size_type],
            tp_value,
        )
        self.computing_data.false_positives_dict[iou_thresh][bbox_size_type] = np.append(
            self.computing_data.false_positives_dict[iou_thresh][bbox_size_type],
            fp_value,
        )

    def __update_computation_dictionaries(
        self,
        bounding_box: BoundingBox,
        iou_thresh: float,
        unmatched_gt_bounding_boxes: List[BoundingBox],
    ) -> List[BoundingBox]:
        bbox_size_type: str = get_bbox_size_type(bounding_box.box)

        # if the ground truth data for this image indicates that nothing has to be detected,
        # indicate this box as False-Positive right away!
        if len(unmatched_gt_bounding_boxes) == 0:
            self.__update_tp_fp_data(
                is_true_positive=False,
                iou_thresh=iou_thresh,
                bbox_size_type=bbox_size_type,
            )
        else:
            max_overlap, assigned_gt_bounding_box = compute_max_bounding_box(
                bounding_box=bounding_box, gt_bounding_boxes=unmatched_gt_bounding_boxes
            )

            # If max-overlap fulfills given threshold
            # the detected box is treated as TP, otherwise as FP
            if (
                max_overlap
                >= iou_thresh
                # and
                # if this bounding_box has not already been assigned as valid,
                # treat it as true positive, otherwise as false positive
                # assigned_bounding_box not in self.detected_annotations[iou_thresh]
            ):
                unmatched_gt_bounding_boxes.remove(assigned_gt_bounding_box)

                # NOTE: In order to have correct metrics, the size of a true positive
                #       bounding box is the same as for the matching ground truth bounding box
                bbox_size_type = get_bbox_size_type(assigned_gt_bounding_box.box)

                # Update as True-Positive
                self.__update_tp_fp_data(
                    is_true_positive=True,
                    iou_thresh=iou_thresh,
                    bbox_size_type=bbox_size_type,
                )
            else:
                # Update as False-Positive
                self.__update_tp_fp_data(
                    is_true_positive=False,
                    iou_thresh=iou_thresh,
                    bbox_size_type=bbox_size_type,
                )

        # append the score of the current bounding-box to the overall  scores list
        self.computing_data.scores[iou_thresh][BBoxSizeTypes.BBOX_ALL] = np.append(
            self.computing_data.scores[iou_thresh][BBoxSizeTypes.BBOX_ALL],
            bounding_box.score,
        )

        # append the score of the current bounding-box to the size specific scores list
        self.computing_data.scores[iou_thresh][bbox_size_type] = np.append(
            self.computing_data.scores[iou_thresh][bbox_size_type], bounding_box.score
        )

        return unmatched_gt_bounding_boxes

    def __update_false_positive_metric_info(
        self,
        ground_truth_annotation: Optional[BaseAnnotation],
        predicted_annotation: BaseAnnotation,
        bounding_box: BoundingBox,
        class_identifier: ClassIdentifier,
    ) -> None:
        # Build an annotation based on the ground truth data and the given
        # bounding boxes
        false_positive_annotation = BaseAnnotation(
            image_path=predicted_annotation.image_path,
            annotation_path=predicted_annotation.annotation_path,
            image_shape=predicted_annotation.image_shape,
            classifications=[],
            bounding_boxes=[bounding_box],
            segmentations=[],
            image_dir=predicted_annotation.image_dir,
            annotation_dir=predicted_annotation.annotation_dir,
            replacement_string=predicted_annotation.replacement_string,
        )

        # save FP annotation to metric_image_info_dict
        if (
            predicted_annotation.image_path
            not in self.model_metrics.metrics_image_info_dict[str(class_identifier)]
        ):
            # The predicted annotation is a FP, therefore store it in a dict so that
            # it can be logged to tensorboard after the evaluation has finished
            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                predicted_annotation.image_path
            ] = MetricImageInfo(
                ground_truth_annotation=ground_truth_annotation,
                false_positive_annotation=false_positive_annotation,
                false_negative_annotation=None,
            )
        else:
            # Update FP annotation object that belongs to the given image-path
            if (
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    predicted_annotation.image_path
                ].false_positive_annotation
                is not None
            ):
                # If their already is an annotation object present, append this bounding box
                # NOTE: the mypy error 'Item "None" of "Optional[BaseAnnotation]" has no
                #       attribute "bounding_boxes' can be ignored. It is checked by the
                #       above query operation. Somehow mypy does not gets that
                self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                    predicted_annotation.image_path
                ].false_positive_annotation.bounding_boxes.append(  # type: ignore
                    bounding_box
                )

            # REMARK: false_positive_annotation of
            #         self.model_metrics.metrics_image_info_dict[class_name]
            #         [predicted_annotation.image_path]
            #         can't be None here. This would only be the case if a MetricImageInfo
            #         object would be inserted with
            #           MetricImageInfo(
            #             ground_truth_annotation=ground_truth_annotation,
            #             false_positive_annotation=None,
            #             false_negative_annotation=false_negative_annotation,
            #           )
            #         Since the false negative annotations will be updated after the false
            #         positive annotations are already set, we always have an
            #         false_positive_annotation present here.

    def __update_false_negative_metric_info(
        self,
        ground_truth_annotation: BaseAnnotation,
        iou_unmatched_gt_bounding_boxes: List[BoundingBox],
        class_identifier: ClassIdentifier,
    ) -> None:
        """
        Update the false_negative_annotation entry of model_metrics.metrics_image_info_dict
        for the given class_name and the image path of the given ground truth annotation.

        Args:
            ground_truth_annotation:
            iou_unmatched_gt_bounding_boxes:
            class_identifier:

        Returns:

        """

        false_negative_annotation = BaseAnnotation(
            image_path=ground_truth_annotation.image_path,
            annotation_path=ground_truth_annotation.annotation_path,
            image_shape=ground_truth_annotation.image_shape,
            classifications=[],
            bounding_boxes=iou_unmatched_gt_bounding_boxes,
            segmentations=[],
            image_dir=ground_truth_annotation.image_dir,
            annotation_dir=ground_truth_annotation.annotation_dir,
            replacement_string=ground_truth_annotation.replacement_string,
        )

        if (
            ground_truth_annotation.image_path
            not in self.model_metrics.metrics_image_info_dict[str(class_identifier)]
        ):
            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                ground_truth_annotation.image_path
            ] = MetricImageInfo(
                ground_truth_annotation=ground_truth_annotation,
                false_positive_annotation=None,
                false_negative_annotation=false_negative_annotation,
            )
        else:
            self.model_metrics.metrics_image_info_dict[str(class_identifier)][
                ground_truth_annotation.image_path
            ].false_negative_annotation = false_negative_annotation

    def __get_gt_annotation(
        self, dataset_index: int, class_identifier: ClassIdentifier
    ) -> Tuple[Optional[BaseAnnotation], List[BoundingBox]]:
        # annotation object for gathering all ground truth bounding boxes for this image
        ground_truth_annotation: Optional[BaseAnnotation] = None

        unmatched_gt_bounding_boxes: List[BoundingBox] = []

        if not str(class_identifier) in self._all_gt_annotations_dict[dataset_index]:
            return ground_truth_annotation, unmatched_gt_bounding_boxes

        # Iterate over all ground-truth annotations that containing bounding-box
        # information for this image (dataset_index) and the given class-id
        for gt_annotation in self._all_gt_annotations_dict[dataset_index][str(class_identifier)]:
            unmatched_gt_bounding_boxes.extend(
                gt_annotation.get_bounding_boxes(include_segmentations=True)
            )

            for bounding_box in gt_annotation.get_bounding_boxes(include_segmentations=True):
                # increase bounding-box ground-truth counter for overall count
                # and the specific bounding-box size count

                self.computing_data.gt_counter_dict[BBoxSizeTypes.BBOX_ALL][
                    str(class_identifier)
                ] += 1

                self.computing_data.gt_counter_dict[get_bbox_size_type(bounding_box.box)][
                    str(class_identifier)
                ] += 1

            # Initialize/update an overall ground_truth annotation that contains all
            # data for this image and class-id
            if ground_truth_annotation is None:
                ground_truth_annotation = copy.deepcopy(gt_annotation)
            else:
                ground_truth_annotation.bounding_boxes.extend(
                    gt_annotation.get_bounding_boxes(include_segmentations=True)
                )

        return ground_truth_annotation, unmatched_gt_bounding_boxes

    def __fill_computing_data(
        self,
        class_id: Optional[int],
        class_name: Optional[str],
        class_identifier: Optional[ClassIdentifier] = None,
    ) -> None:
        if class_identifier is None:
            if class_id is not None:
                logger.warning(
                    "DEPRECATED: The parameter class_id is deprecated "
                    "and will be removed in future versions, please use the parameter "
                    "'class_identifier' instead"
                )

                if class_name is not None:
                    logger.warning(
                        "DEPRECATED: The parameter class_id is deprecated "
                        "and will be removed in future versions, please use the parameter "
                        "'class_identifier' instead"
                    )
                else:
                    class_name = self.mapper.map_annotation_class_id_to_model_class_name(
                        class_id=class_id
                    )
            else:
                if class_name is not None:
                    logger.warning(
                        "DEPRECATED: The parameter class_id is deprecated "
                        "and will be removed in future versions, please use the parameter "
                        "'class_identifier' instead"
                    )

                    class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                        class_name=class_name
                    )
                else:
                    raise ValueError("TODO")

            class_identifier = ClassIdentifier(class_id=class_id, class_name=class_name)

        process_bar = tqdm(
            range(self.dataset_length),
            desc=f"Compute metrics for class-identifier: {class_identifier}",
        )

        # Iterate over all image indices
        for dataset_index in process_bar:
            (ground_truth_annotation, unmatched_gt_bounding_boxes,) = self.__get_gt_annotation(
                dataset_index=dataset_index, class_identifier=class_identifier
            )

            for iou_thresh in self.iou_thresholds:
                # Iterate over all predicted annotations for this image and class-id
                iou_unmatched_gt_bounding_boxes = copy.deepcopy(unmatched_gt_bounding_boxes)

                if str(class_identifier) in self._all_predicted_annotations_dict[dataset_index]:
                    for predicted_annotation in self._all_predicted_annotations_dict[
                        dataset_index
                    ][str(class_identifier)]:
                        for bounding_box in predicted_annotation.get_bounding_boxes(
                            include_segmentations=True
                        ):
                            iou_unmatched_gt_bounding_boxes = (
                                self.__update_computation_dictionaries(
                                    bounding_box=bounding_box,
                                    iou_thresh=iou_thresh,
                                    unmatched_gt_bounding_boxes=iou_unmatched_gt_bounding_boxes,
                                )
                            )

                            # The bounding_box is a FP
                            if self.computing_data.false_positives_dict[iou_thresh][
                                BBoxSizeTypes.BBOX_ALL
                            ][-1]:
                                self.__update_false_positive_metric_info(
                                    ground_truth_annotation=ground_truth_annotation,
                                    predicted_annotation=predicted_annotation,
                                    bounding_box=bounding_box,
                                    class_identifier=class_identifier,
                                )

                    # There are ground truth annotations which haven't been matched,
                    # this states that the box is a false negative
                    if (
                        str(class_identifier) in self._all_gt_annotations_dict[dataset_index]
                        and len(
                            self._all_gt_annotations_dict[dataset_index][str(class_identifier)]
                        )
                        > 0
                        and len(iou_unmatched_gt_bounding_boxes) > 0
                        and ground_truth_annotation is not None
                    ):
                        self.__update_false_negative_metric_info(
                            ground_truth_annotation=ground_truth_annotation,
                            iou_unmatched_gt_bounding_boxes=iou_unmatched_gt_bounding_boxes,
                            class_identifier=class_identifier,
                        )

    def __compute_and_update_metrics_step(
        self,
        class_identifier: ClassIdentifier,
        iou_thresh: float,
        bbox_size_type: str,
        num_annotations: int,
        true_positives: np.ndarray,  # type: ignore[type-arg]
        false_positives: np.ndarray,  # type: ignore[type-arg]
    ) -> None:
        # compute recall and precision
        recall_values = true_positives / num_annotations
        precision_values = true_positives / np.maximum(
            true_positives + false_positives, np.finfo(np.float64).eps
        )

        # compute average precision
        # NOTE: mypy error 'Call to untyped function "close" in typed context' can be ignored
        ap = voc_ap(rec=recall_values, prec=precision_values, use_07_metric=False)  # type: ignore

        if len(true_positives) > 0:
            tp = int(true_positives[-1])

            if tp > num_annotations:
                print("TP not valid")

            rc = tp / num_annotations
        else:
            tp = DEFAULT_INT_VALUE
            rc = DEFAULT_FLOAT_VALUE

        if len(false_positives) > 0:
            fp = int(false_positives[-1])
        else:
            fp = DEFAULT_INT_VALUE

        try:
            pr = tp / (fp + tp)
        except ZeroDivisionError:
            pr = DEFAULT_FLOAT_VALUE

        try:
            f1 = 2 * (pr * rc) / (pr + rc)
        except ZeroDivisionError:
            f1 = DEFAULT_FLOAT_VALUE

        fn = num_annotations - tp

        self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][
            str(class_identifier)
        ] = ODMetrics(TP=tp, FP=fp, FN=fn, PR=pr, RC=rc, F1=f1, AP=ap, COUNT=num_annotations)

    def __compute_metrics(self, class_identifier: ClassIdentifier) -> None:
        # TODO: Is this mAP computation? - separate from rest !
        for iou_thresh in self.iou_thresholds:
            for bbox_size_type in BBoxSizeTypes.get_values_as_list(class_type=BBoxSizeTypes):
                num_annotations = self.computing_data.gt_counter_dict[bbox_size_type][
                    str(class_identifier)
                ]

                if num_annotations == 0.0:
                    continue

                # sort by score
                # TODO: CHECK NEGATIVE SIGN!!!
                indices = np.argsort(-self.computing_data.scores[iou_thresh][bbox_size_type])

                false_positives = self.computing_data.false_positives_dict[iou_thresh][
                    bbox_size_type
                ][indices]

                true_positives = self.computing_data.true_positives_dict[iou_thresh][
                    bbox_size_type
                ][indices]

                # compute false positives and true positives
                false_positives = np.cumsum(false_positives)
                true_positives = np.cumsum(true_positives)

                self.__compute_and_update_metrics_step(
                    class_identifier=class_identifier,
                    iou_thresh=iou_thresh,
                    bbox_size_type=bbox_size_type,
                    num_annotations=num_annotations,
                    true_positives=true_positives,
                    false_positives=false_positives,
                )

                if (
                    self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][
                        str(class_identifier)
                    ].FP
                    != -1
                    and self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][
                        str(class_identifier)
                    ].TP
                    != -1
                ):
                    self.computing_data.valid_precisions[iou_thresh][bbox_size_type].append(
                        self.model_metrics.metrics_dict[iou_thresh][bbox_size_type][
                            str(class_identifier)
                        ].AP
                    )

    @staticmethod
    def match_false_negatives_and_false_positives(
        metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE,
        iou_threshold: float,
    ) -> Tuple[METRIC_IMAGE_INFO_TYPE, CONFUSION_MATRIX_TYPE]:
        """


        Args:
            metrics_image_info_dict:
            iou_threshold:

        Returns:

        """

        logger.warning(
            "DEPRECATED: The method 'match_false_negatives_and_false_positives' is "
            "deprecated and will be removed in future versions"
        )

        confusion_matrix: CONFUSION_MATRIX_TYPE = [
            [0 for _ in metrics_image_info_dict.keys()] for _ in metrics_image_info_dict.keys()
        ]
        classes_name_dict: Dict[str, int] = {}
        for class_identifier_str in metrics_image_info_dict.keys():
            class_identifier: ClassIdentifier = ClassIdentifier.from_str(
                class_identifier_str=class_identifier_str
            )
            classes_name_dict[str(class_identifier)] = class_identifier.class_id

        # Build a dict for reverse look up, where the keys are image paths.
        # This is needed to fill the confusion matrix per image.
        image_metrics_info_dict: Dict[str, Dict[str, MetricImageInfo]] = {}
        for class_identifier_str in metrics_image_info_dict.keys():
            for image_path, metric_image_info in metrics_image_info_dict[
                class_identifier_str
            ].items():
                if image_path not in image_metrics_info_dict:
                    image_metrics_info_dict[image_path] = {}

                if class_identifier_str not in image_metrics_info_dict[image_path]:
                    image_metrics_info_dict[image_path][class_identifier_str] = metric_image_info

        result_metrics_image_info_dict = copy.deepcopy(metrics_image_info_dict)

        for image_path in image_metrics_info_dict.keys():
            class_identifiers = [
                ClassIdentifier.from_str(class_identifier_str=class_identifier_str)
                for class_identifier_str in image_metrics_info_dict[image_path].keys()
            ]

            for class_identifier in class_identifiers:
                other_class_identifiers = copy.deepcopy(class_identifiers)
                other_class_identifiers.remove(class_identifier)

                class_img_metrics_info = image_metrics_info_dict[image_path][str(class_identifier)]
                if class_img_metrics_info.false_negative_annotation:
                    class_fn_bounding_boxes = (
                        class_img_metrics_info.false_negative_annotation.bounding_boxes
                    )

                    for fn_bounding_box in class_fn_bounding_boxes:
                        for other_class_identifier in other_class_identifiers:
                            other_class_img_metrics_info = image_metrics_info_dict[image_path][
                                str(other_class_identifier)
                            ]

                            if other_class_img_metrics_info.false_positive_annotation:
                                other_class_fp_bounding_boxes = (
                                    other_class_img_metrics_info.false_positive_annotation.bounding_boxes
                                )

                                (
                                    max_overlap,
                                    assigned_fp_bounding_box,
                                ) = compute_max_bounding_box(
                                    bounding_box=fn_bounding_box,
                                    gt_bounding_boxes=other_class_fp_bounding_boxes,
                                )

                                if max_overlap > iou_threshold:
                                    confusion_matrix[classes_name_dict[str(class_identifier)]][
                                        classes_name_dict[str(other_class_identifier)]
                                    ] += 1
                                    annotation_path = (
                                        class_img_metrics_info.false_negative_annotation.annotation_path
                                    )
                                    image_shape = (
                                        class_img_metrics_info.false_negative_annotation.image_shape
                                    )
                                    image_dir = (
                                        class_img_metrics_info.false_negative_annotation.image_dir
                                    )
                                    annotation_dir = (
                                        class_img_metrics_info.false_negative_annotation.annotation_dir
                                    )
                                    replacement_string = (
                                        class_img_metrics_info.false_negative_annotation.replacement_string
                                    )

                                    class_img_metrics_info.false_negative_annotation.bounding_boxes.remove(
                                        fn_bounding_box
                                    )
                                    other_class_img_metrics_info.false_positive_annotation.bounding_boxes.remove(
                                        assigned_fp_bounding_box
                                    )

                                    result_metrics_image_info_dict = MetricsComputation.__change_matched_bbox_attribute(
                                        metrics_image_info_dict=result_metrics_image_info_dict,
                                        image_path=image_path,
                                        annotation_path=annotation_path,
                                        class_identifier=class_identifier,
                                        other_class_identifier=other_class_identifier,
                                        image_shape=image_shape,
                                        image_dir=image_dir,
                                        annotation_dir=annotation_dir,
                                        replacement_string=replacement_string,
                                        fn_bounding_box=fn_bounding_box,
                                        assigned_fp_bounding_box=assigned_fp_bounding_box,
                                        class_img_metrics_info=class_img_metrics_info,
                                        other_class_img_metrics_info=other_class_img_metrics_info,
                                    )

                            if other_class_img_metrics_info.false_positive_annotation:
                                if (
                                    len(
                                        other_class_img_metrics_info.false_positive_annotation.bounding_boxes
                                    )
                                    == 0
                                ):
                                    other_class_img_metrics_info.false_positive_annotation = None

                if class_img_metrics_info.false_negative_annotation:
                    if len(class_img_metrics_info.false_negative_annotation.bounding_boxes) == 0:
                        class_img_metrics_info.false_negative_annotation = None

        return result_metrics_image_info_dict, confusion_matrix

    @staticmethod
    def match_false_negatives_and_false_positives_as_dict(
        metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE,
        iou_threshold: float,
    ) -> Tuple[METRIC_IMAGE_INFO_TYPE, CONFUSION_MATRIX_DICT_TYPE]:
        """


        Args:
            metrics_image_info_dict:
            iou_threshold:

        Returns:

        """

        confusion_matrix: CONFUSION_MATRIX_DICT_TYPE = {
            k2: {k1: 0 for k1 in metrics_image_info_dict.keys()}
            for k2 in metrics_image_info_dict.keys()
        }

        # Build a dict for reverse look up (reverse of metrics_image_info_dict),
        # where the keys are image paths. This is needed to fill the confusion matrix per image.
        # 1st key: Image Path
        # 2nd key: Class Identifier as string
        # value: The MetricImageInfo for this image path and class name
        image_metrics_info_dict: Dict[str, Dict[str, MetricImageInfo]] = {}
        for class_identifier_str in metrics_image_info_dict.keys():
            for image_path, metric_image_info in metrics_image_info_dict[
                class_identifier_str
            ].items():
                if image_path not in image_metrics_info_dict:
                    image_metrics_info_dict[image_path] = {}

                if class_identifier_str not in image_metrics_info_dict[image_path]:
                    image_metrics_info_dict[image_path][class_identifier_str] = metric_image_info

        result_metrics_image_info_dict = copy.deepcopy(metrics_image_info_dict)

        for image_path in image_metrics_info_dict.keys():
            class_identifiers: List[ClassIdentifier] = [
                ClassIdentifier.from_str(class_identifier_str=c)
                for c in image_metrics_info_dict[image_path].keys()
            ]

            for class_identifier in class_identifiers:
                other_class_identifiers: List[ClassIdentifier] = copy.deepcopy(class_identifiers)
                other_class_identifiers.remove(class_identifier)

                class_img_metrics_info = image_metrics_info_dict[image_path][str(class_identifier)]
                if class_img_metrics_info.false_negative_annotation:
                    class_fn_bounding_boxes = (
                        class_img_metrics_info.false_negative_annotation.bounding_boxes
                    )

                    for fn_bounding_box in class_fn_bounding_boxes:
                        for other_class_identifier in other_class_identifiers:
                            other_class_img_metrics_info = image_metrics_info_dict[image_path][
                                str(other_class_identifier)
                            ]

                            if other_class_img_metrics_info.false_positive_annotation:
                                other_class_fp_bounding_boxes = (
                                    other_class_img_metrics_info.false_positive_annotation.bounding_boxes
                                )

                                (
                                    max_overlap,
                                    assigned_fp_bounding_box,
                                ) = compute_max_bounding_box(
                                    bounding_box=fn_bounding_box,
                                    gt_bounding_boxes=other_class_fp_bounding_boxes,
                                )

                                if max_overlap > iou_threshold:
                                    confusion_matrix[str(class_identifier)][
                                        str(other_class_identifier)
                                    ] += 1
                                    annotation_path = (
                                        class_img_metrics_info.false_negative_annotation.annotation_path
                                    )
                                    image_shape = (
                                        class_img_metrics_info.false_negative_annotation.image_shape
                                    )
                                    image_dir = (
                                        class_img_metrics_info.false_negative_annotation.image_dir
                                    )
                                    annotation_dir = (
                                        class_img_metrics_info.false_negative_annotation.annotation_dir
                                    )
                                    replacement_string = (
                                        class_img_metrics_info.false_negative_annotation.replacement_string
                                    )

                                    class_img_metrics_info.false_negative_annotation.bounding_boxes.remove(
                                        fn_bounding_box
                                    )
                                    other_class_img_metrics_info.false_positive_annotation.bounding_boxes.remove(
                                        assigned_fp_bounding_box
                                    )

                                    result_metrics_image_info_dict = MetricsComputation.__change_matched_bbox_attribute(
                                        metrics_image_info_dict=result_metrics_image_info_dict,
                                        image_path=image_path,
                                        annotation_path=annotation_path,
                                        class_identifier=class_identifier,
                                        other_class_identifier=other_class_identifier,
                                        image_shape=image_shape,
                                        image_dir=image_dir,
                                        annotation_dir=annotation_dir,
                                        replacement_string=replacement_string,
                                        fn_bounding_box=fn_bounding_box,
                                        assigned_fp_bounding_box=assigned_fp_bounding_box,
                                        class_img_metrics_info=class_img_metrics_info,
                                        other_class_img_metrics_info=other_class_img_metrics_info,
                                    )

                            if other_class_img_metrics_info.false_positive_annotation:
                                if (
                                    len(
                                        other_class_img_metrics_info.false_positive_annotation.bounding_boxes
                                    )
                                    == 0
                                ):
                                    other_class_img_metrics_info.false_positive_annotation = None

                if class_img_metrics_info.false_negative_annotation:
                    if len(class_img_metrics_info.false_negative_annotation.bounding_boxes) == 0:
                        class_img_metrics_info.false_negative_annotation = None

        return result_metrics_image_info_dict, confusion_matrix

    @staticmethod
    def __change_matched_bbox_attribute(
        metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE,
        image_path: str,
        annotation_path: str,
        class_identifier: ClassIdentifier,
        other_class_identifier: ClassIdentifier,
        image_shape: Tuple[int, int],
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
        fn_bounding_box: BoundingBox,
        assigned_fp_bounding_box: BoundingBox,
        class_img_metrics_info: MetricImageInfo,
        other_class_img_metrics_info: MetricImageInfo,
    ) -> METRIC_IMAGE_INFO_TYPE:
        if class_img_metrics_info.false_negative_matched_false_positive_annotation is not None:
            class_img_metrics_info.false_negative_matched_false_positive_annotation.bounding_boxes.append(
                fn_bounding_box
            )
        else:
            class_img_metrics_info.false_negative_matched_false_positive_annotation = (
                BaseAnnotation(
                    image_path=image_path,
                    annotation_path=annotation_path,
                    image_shape=image_shape,
                    classifications=[],
                    bounding_boxes=[fn_bounding_box],
                    segmentations=[],
                    image_dir=image_dir,
                    annotation_dir=annotation_dir,
                    replacement_string=replacement_string,
                )
            )

        if (
            other_class_img_metrics_info.false_negative_matched_false_positive_annotation
            is not None
        ):
            other_class_img_metrics_info.false_negative_matched_false_positive_annotation.bounding_boxes.append(
                assigned_fp_bounding_box
            )

        else:
            other_class_img_metrics_info.false_negative_matched_false_positive_annotation = (
                BaseAnnotation(
                    image_path=image_path,
                    annotation_path=annotation_path,
                    image_shape=image_shape,
                    classifications=[],
                    bounding_boxes=[assigned_fp_bounding_box],
                    segmentations=[],
                    image_dir=image_dir,
                    annotation_dir=annotation_dir,
                    replacement_string=replacement_string,
                )
            )

        metrics_image_info_dict[str(class_identifier)][image_path] = class_img_metrics_info
        metrics_image_info_dict[str(other_class_identifier)][
            image_path
        ] = other_class_img_metrics_info

        return metrics_image_info_dict



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for building COCO formatted annotations."""

import logging
from typing import Any, Dict, List, Tuple, cast

import numpy as np

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_builder import AnnotationBuilder
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.segmentation import PolygonType, Segmentation
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.data_preparation.utils import ensure_abspath
from mlcvzoo_base.third_party.imutils.perspective import order_points

logger = logging.getLogger(__name__)


class COCOAnnotationBuilder(AnnotationBuilder):
    """
    Super class for defining the methods that are needed to build a BaseAnnotation
    object from a COCO type XML file.
    """

    def __init__(
        self,
        image_shape: Tuple[int, int],  # HEIGHT, WIDTH
        image_id: str,
        coco_annotations: List[Dict[str, Any]],
        coco_categories: List[Dict[str, Any]],
        mapper: AnnotationClassMapper,
        use_difficult: bool,
    ) -> None:
        AnnotationBuilder.__init__(self)

        self.image_shape = image_shape
        self.image_id = image_id
        self.coco_annotations = coco_annotations
        self.mapper = mapper

        self.coco_classes_id_dict: Dict[int, str] = self.mapper.get_coco_classes_id_dict(
            categories=coco_categories
        )

        self.use_difficult = use_difficult

    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        bounding_boxes, segmentations = self.__get_coco_annotations()

        annotation: BaseAnnotation = BaseAnnotation(
            image_path=image_path,
            annotation_path=annotation_path,
            image_shape=self.image_shape,
            classifications=[],
            bounding_boxes=bounding_boxes,
            segmentations=segmentations,
            image_dir=image_dir,
            annotation_dir=annotation_dir,
            replacement_string=replacement_string,
        )

        try:
            AnnotationBuilder._check_annotation(annotation=annotation)
        except ValueError as value_error:
            logger.exception(
                f"{value_error}, in a future version, the whole annotation will be skipped!"
            )

        annotation = ensure_abspath(annotation=annotation)

        return annotation

    @staticmethod
    def __get_attributes_from_coco_annotation(
        coco_annotation: Dict[str, Any]
    ) -> Tuple[bool, bool, str]:
        """
        Sets multiple attributes which characterize the corresponding bounding box

        Args:
            coco_annotation: Dict containing information about the bounding box

        Returns:
            Tuple containing three attributes

        """

        if "attributes" in coco_annotation:
            attributes = coco_annotation["attributes"]

            difficult = attributes["difficult"] if "difficult" in attributes else False
            occluded = attributes["occluded"] if "occluded" in attributes else False
            content = attributes["content"] if "content" in attributes else ""
        else:
            difficult = False
            occluded = False
            content = ""

        return difficult, occluded, content

    @staticmethod
    def __create_polygon_from_coco_polygon(
        coco_polygon: List[List[float]],
    ) -> PolygonType:
        """
        Converts a polygon given from an annotation file in COCO formats
        to a MLCVZoo conform polygon of type PolygonType

        Args:
            coco_polygon: Polygon in COCO format

        Returns:
            Polygon of PolygonType
        """

        main_coco_polygon: List[float] = coco_polygon[0]

        polygon_tuple_list: PolygonType = []
        for index in range(0, len(main_coco_polygon), 2):
            new_point = (main_coco_polygon[index], main_coco_polygon[index + 1])
            polygon_tuple_list.append(new_point)

        if len(polygon_tuple_list) == 4:
            polygon_tuple_list = [
                cast(Tuple[float, float], tuple(point))
                for point in order_points(
                    polygon=np.array(polygon_tuple_list, dtype="float32"),
                    sort_by_euclidean=False,
                )
            ]

        return polygon_tuple_list

    def __get_coco_annotations(
        self,
    ) -> Tuple[List[BoundingBox], List[Segmentation]]:
        """
        Documentation regarding the COCO format can be found here:
        https://cocodataset.org/#format-data

        """

        _bounding_boxes: List[BoundingBox] = list()
        _segmentations: List[Segmentation] = list()

        for _coco_annotation in self.coco_annotations:
            if _coco_annotation["image_id"] != self.image_id:
                continue

            # Get the coco class name based on the given coco category_id
            coco_class_name = self.coco_classes_id_dict[int(_coco_annotation["category_id"])]

            try:
                # map the parsed "class_name" according to the mapping defined in the mapper class
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=coco_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=coco_class_name
                )
            except ClassMappingNotFoundError:
                logger.warning(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "BndBox will be skipped, coco-annotation= '%s'"
                    % (coco_class_name, _coco_annotation)
                )
                continue

            (
                difficult,
                occluded,
                content,
            ) = self.__get_attributes_from_coco_annotation(coco_annotation=_coco_annotation)

            box = Box.init_format_based(
                box_list=_coco_annotation["bbox"],
                box_format=ObjectDetectionBBoxFormats.XYWH,
                src_shape=self.image_shape,
            )

            # NOTE: COCO bbox is format XYWH
            bounding_box = BoundingBox(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                model_class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=0.0,
                difficult=difficult,
                occluded=occluded,
                content=content,
                box=box,
            )

            # TODO: handle occluded as well?
            if difficult and not self.use_difficult:
                logger.debug("Found difficult box, which will be skipped! %s", bounding_box)
                continue

            coco_polygon = _coco_annotation["segmentation"]
            if len(coco_polygon) > 0:
                polygon = self.__create_polygon_from_coco_polygon(coco_polygon=coco_polygon)

                segmentation = Segmentation(
                    class_identifier=ClassIdentifier(
                        class_id=class_id,
                        class_name=class_name,
                    ),
                    model_class_identifier=ClassIdentifier(
                        class_id=class_id,
                        class_name=class_name,
                    ),
                    score=0.0,
                    difficult=difficult,
                    occluded=occluded,
                    content=content,
                    polygon=polygon,
                    box=box,
                )

                # Append segmentation given with surrounding box
                _segmentations.append(segmentation)
            else:
                # No segmentation given. The bounding-box is a real bounding-box.
                _bounding_boxes.append(bounding_box)

        return _bounding_boxes, _segmentations



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing CVAT formatted annotations"""
import logging
import xml.etree.ElementTree as ET_xml
from typing import List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError, ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerSingleFileInputDataConfig,
)
from mlcvzoo_base.data_preparation.annotation_builder.cvat_annotation_builder import (
    CVATAnnotationBuilder,
)

logger = logging.getLogger(__name__)


class CVATAnnotationParser(AnnotationParser):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        cvat_input_data: List[AnnotationHandlerSingleFileInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.cvat_input_data = cvat_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = []

        for dataset_count, input_data in enumerate(self.cvat_input_data):
            tree = ET_xml.parse(input_data.input_path)
            root = tree.getroot()

            if root is not None:
                images = root.findall("image")

                for image in images:
                    cvat_tags = image.findall("tag")
                    cvat_bboxes = image.findall("box")
                    cvat_polygons = image.findall("polygon")

                    try:
                        classifications = self.__parse_cvat_classifications(cvat_tags=cvat_tags)
                        bounding_boxes = self.__parse_cvat_bounding_boxes(
                            cvat_bboxes=cvat_bboxes,
                            use_difficult=input_data.use_difficult,
                        )
                        segmentations = self.__parse_cvat_polygons(
                            cvat_polygons=cvat_polygons,
                            use_difficult=input_data.use_difficult,
                        )

                        image_path = image.attrib["name"]
                        image_width = int(image.attrib["width"])
                        image_height = int(image.attrib["height"])

                        replacement_string = (
                            AnnotationParser.csv_directory_replacement_string.format(dataset_count)
                        )

                        cvat_builder = CVATAnnotationBuilder(
                            image_shape=(image_height, image_width),
                            input_classifications=classifications,
                            input_bounding_boxes=bounding_boxes,
                            input_segmentations=segmentations,
                        )

                        annotation = cvat_builder.build(
                            image_path=image_path,
                            annotation_path=input_data.input_path,
                            image_dir=input_data.input_root_dir,
                            annotation_dir=input_data.input_root_dir,
                            replacement_string=replacement_string,
                        )
                        annotations.append(annotation)

                    except (ValueError, ForbiddenClassError) as error:
                        logger.warning("%s, annotation will be skipped" % str(error))
                        continue

        return annotations

    def __parse_cvat_classifications(
        self, cvat_tags: List[ET_xml.Element]
    ) -> List[Classification]:
        classifications: List[Classification] = []

        for tag in cvat_tags:
            xml_class_name = tag.attrib["label"]

            try:
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=xml_class_name
                )
                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=xml_class_name
                )
            except ClassMappingNotFoundError:
                logger.warning(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "Classification will be skipped" % xml_class_name
                )
                continue

            classification = Classification(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                model_class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=1.0,
            )

            classifications.append(classification)

        return classifications

    def __parse_cvat_bounding_boxes(
        self, cvat_bboxes: List[ET_xml.Element], use_difficult: bool
    ) -> List[BoundingBox]:
        bounding_boxes: List[BoundingBox] = []

        for cvat_bbox in cvat_bboxes:
            xml_class_name = cvat_bbox.attrib["label"]

            try:
                # map the parsed "class_name" according to the mapping defined in the mapper class
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=xml_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=xml_class_name
                )
            except ClassMappingNotFoundError:
                logger.warning(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "BndBox will be skipped, cvat-box= '%s'"
                    % (
                        xml_class_name,
                        cvat_bbox,
                    )
                )
                continue

            occluded, difficult, content = CVATAnnotationParser.__read_attributes(
                cvat_bbox.findall("attribute")
            )

            bounding_box = BoundingBox(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                model_class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=1.0,
                box=Box(
                    xmin=int(float(cvat_bbox.attrib["xtl"])),
                    ymin=int(float(cvat_bbox.attrib["ytl"])),
                    xmax=int(float(cvat_bbox.attrib["xbr"])),
                    ymax=int(float(cvat_bbox.attrib["ybr"])),
                ),
                occluded=occluded,
                difficult=difficult,
                content=content,
            )

            # TODO: handle occluded as well?
            if difficult and not use_difficult:
                logger.debug("Found difficult box, which will be skipped! %s", bounding_box)
                continue

            bounding_boxes.append(bounding_box)

        return bounding_boxes

    def __parse_cvat_polygons(
        self, cvat_polygons: List[ET_xml.Element], use_difficult: bool
    ) -> List[Segmentation]:
        segmentations: List[Segmentation] = []

        for cvat_polygon in cvat_polygons:
            xml_class_name = cvat_polygon.attrib["label"]

            try:
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=xml_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=xml_class_name
                )
            except ClassMappingNotFoundError:
                logger.warning(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "Segmentation will be skipped, cvat-segmentation= '%s'",
                    xml_class_name,
                    cvat_polygon,
                )
                continue

            cvat_points = cvat_polygon.attrib["points"].split(";")

            polygon: List[Tuple[float, float]] = []

            for cvat_point in cvat_points:
                point_split = cvat_point.split(",")

                x = float(point_split[0])
                y = float(point_split[1])

                polygon.append((x, y))

            occluded, difficult, content = CVATAnnotationParser.__read_attributes(
                cvat_polygon.findall("attribute")
            )

            segmentation = Segmentation(
                class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                model_class_identifier=ClassIdentifier(
                    class_id=class_id,
                    class_name=class_name,
                ),
                score=1.0,
                polygon=polygon,
                occluded=occluded,
                difficult=difficult,
                content=content,
            )

            # TODO: handle occluded as well?
            if difficult and not use_difficult:
                logger.debug(
                    "Found difficult segmentation, which will be skipped! %s",
                    segmentation,
                )
                continue

            segmentations.append(segmentation)

        return segmentations

    @staticmethod
    def __read_attributes(attributes: List[ET_xml.Element]) -> Tuple[bool, bool, str]:
        occluded = False
        difficult = False
        content = ""

        for attribute in attributes:
            if attribute.attrib["name"] == "occluded":
                occluded = False if attribute.text == "false" else True
            if attribute.attrib["name"] == "difficult":
                difficult = False if attribute.text == "false" else True
            if attribute.attrib["name"] == "content":
                content = str(attribute.text)

        return occluded, difficult, content

    def parse_cvat_meta_info(self) -> List[ET_xml.Element]:
        """
        Parses all tags under meta tag into a list of XML Elements.

        Returns: a list of XML Elements

        """

        meta_elements: List[ET_xml.Element] = []
        for dataset_count, input_data in enumerate(self.cvat_input_data):
            tree = ET_xml.parse(input_data.input_path)
            root = tree.getroot()

            if root is not None:
                meta_tags = root.findall("meta")

                if len(meta_tags) > 1:
                    logger.warning(
                        "Found more than one <meta> tag in given XML file. "
                        "Format is not as expected."
                    )
                meta_elements.append(meta_tags[0])

        return meta_elements



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for using MLFlow tool in MLCVZoo"""
import copy
import logging
import re
import typing
from typing import Dict, List, Optional

import mlflow
from config_builder import ConfigBuilder
from mlflow import ActiveRun
from mlflow.entities import Experiment

from mlcvzoo_base.configuration.mlfow_config import MLFlowConfig
from mlcvzoo_base.configuration.structs import (
    MLFlowExperimentConfig,
    MLFLowTrackingUriTypes,
)

logger = logging.getLogger(__name__)


class MLFLowRunner:
    """
    Class for wrapping the start and stop of mlflow runs
    """

    def __init__(
        self,
        configuration: Optional[MLFlowConfig] = None,
        yaml_config_path: Optional[str] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
        create_experiments: bool = True,
    ):
        """
        Init with all parameters that can be set for the config-builder.

        Args:
            configuration:
            yaml_config_path:
            string_replacement_map:
            no_checks:
            create_experiments:

        Returns:
            The created object
        """

        self.mlflow_run: Optional[ActiveRun] = None
        self.artifact_location: Optional[str] = None

        if configuration is not None:
            self.configuration: MLFlowConfig = configuration
        else:
            # Build configuration by instantiating a config-builder object
            config_builder = ConfigBuilder(
                class_type=MLFlowConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            )

            self.configuration = typing.cast(
                MLFlowConfig, copy.deepcopy(config_builder.configuration)
            )

        # Setup base experiments
        if create_experiments:
            self.create_mlflow_experiments()

    def create_mlflow_experiments(self) -> None:
        """
        Create all mlflow experiments that are specified in MLFlowExperimentConfig.EXPERIMENT_DICT.
        Init according to the configured backed
        -> mlflow_file_config        => setup mlflow to write data to local directory
        -> mlflow_postgressql_config => setup mlflow to write to postgresql database

        :return:
        """

        if self.configuration.mlflow_file_config is not None:
            mlflow.set_tracking_uri(
                f"{MLFLowTrackingUriTypes.FILE}:{self.configuration.mlflow_file_config.logging_dir}"
            )

        elif self.configuration.mlflow_postgressql_config is not None:
            DB_USER = self.configuration.mlflow_postgressql_config.database_user
            DB_PW = self.configuration.mlflow_postgressql_config.database_pw
            DB_PORT = self.configuration.mlflow_postgressql_config.database_port
            DB_NAME = self.configuration.mlflow_postgressql_config.database_name

            mlflow.set_tracking_uri(
                f"{MLFLowTrackingUriTypes.POSTGRES}://{DB_USER}:{DB_PW}@localhost:{DB_PORT}/{DB_NAME}"
            )
        else:
            logger.error("No valid configuration given, could not create an mlflow experiment")
            return

        for (
            experiment_id,
            experiment_name,
        ) in MLFlowExperimentConfig.EXPERIMENT_DICT.items():
            try:
                mlflow_experiment_id = mlflow.create_experiment(
                    name=experiment_name,
                    artifact_location=self.configuration.artifact_location,
                )

                logger.info(
                    f"Created Experiment: "
                    f"ID={mlflow_experiment_id}, "
                    f"NAME={experiment_name}, "
                    f"ARTIFACT_LOCATION={self.configuration.artifact_location}"
                )
            except mlflow.exceptions.MlflowException as e:
                reg_search = re.search(
                    pattern="already exist",
                    string=str(e),
                    flags=re.M | re.I,
                )

                if reg_search is not None:
                    logger.debug(f"Ignore MlflowException: {str(e)}")
                else:
                    logger.exception(e)

    @staticmethod
    def end_run() -> None:
        """
        Terminate mlflow run (if there is one).
        Catch known errors
        """
        try:
            mlflow.end_run()
        except mlflow.exceptions.MlflowException as mlflow_exception:
            logger.warning(mlflow_exception)

    def start_mlflow_run(
        self,
        run_id: Optional[str] = None,
        experiment_name: Optional[str] = None,
        run_name: Optional[str] = None,
        end_runs_in_advance: bool = True,
    ) -> Optional[str]:
        """
        Start an mlflow run. Two options:
        1) By run-id, this will continue the run with the specified ID
        2) By experiment- and run-name

        :param experiment_name:
        :param run_name:
        :param run_id:
        :param end_runs_in_advance:
        :return:
        """

        # Ensure that no other mlflow run is started
        if end_runs_in_advance:
            MLFLowRunner.end_run()

        # Use run-id and continue mlflow run
        if run_id is not None:
            try:
                self.mlflow_run = mlflow.start_run(run_id=run_id)

                logger.info(
                    "=========================================\n"
                    f"Continue mlflow run:\n{self.mlflow_run}\n"
                )

                return str(self.mlflow_run.info.run_id)

            except mlflow.exceptions.MlflowException:
                logger.warning(
                    f"Could not continue mlflow run with run-id {run_id}. Start new mlflow run ..."
                )

        # start new mlflow run with given experiment_name and run_name
        elif experiment_name is not None and run_name is not None:
            mlflow_experiment: Optional[Experiment]

            if experiment_name in MLFlowExperimentConfig.EXPERIMENT_DICT.values():
                mlflow_experiment = mlflow.get_experiment_by_name(name=experiment_name)
            else:
                logger.error(
                    "experiment-name is not valid. "
                    f"Provide one of {MLFlowExperimentConfig.EXPERIMENT_DICT.values()}"
                )
                return None

            if mlflow_experiment is not None:
                self.mlflow_run = mlflow.start_run(
                    experiment_id=mlflow_experiment.experiment_id, run_name=run_name
                )

                logger.info(
                    "=========================================\n"
                    f"New mlflow run: \n{self.mlflow_run.info}\n"
                )

                return str(self.mlflow_run.info.run_id)
            else:
                logger.error(
                    f"Could not start a mlflow run for experiment-name '{experiment_name}'"
                )

        return None



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for configuring the parsing of information from yaml in python
accessible attributes for the ReadFromFileModel class
"""
from typing import Optional

import related
from attr import define

from mlcvzoo_base.api.configuration import ModelConfiguration
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import ReductionMappingConfig


@define
class ReadFromFileConfig(ModelConfiguration):
    """
    Class for parsing information from yaml in respective hierarchy

    Attributes:
        class_mapping (ClassMappingConfig):
        annotation_handler_config:
        use_image_name_hash:
        include_segmentations:
        reduction_class_mapping:
    """

    class_mapping: ClassMappingConfig = related.ChildField(ClassMappingConfig)

    annotation_handler_config: AnnotationHandlerConfig = related.ChildField(
        cls=AnnotationHandlerConfig
    )

    use_image_name_hash: bool = related.BooleanField(default=False)

    include_segmentations: bool = related.BooleanField(default=False)

    reduction_class_mapping: Optional[ReductionMappingConfig] = related.ChildField(
        cls=ReductionMappingConfig, required=False, default=None
    )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations regarding files and file access"""
import logging
import os
import shutil
import zipfile
from pathlib import Path
from typing import List, Tuple

import cv2
import numpy as np

logger = logging.getLogger(__name__)


def get_file_list(
    input_dir: str,
    search_subfolders: bool = False,
    file_extension: str = "",
    include_pattern: str = "",
    exclude_pattern: str = "",
) -> List[str]:
    """
    Create a list of all files found in input_dir. When search_subfolders = True, the list
    contains all files from the input_dir and all its subfolders. The file_extension argument
    can be set to filter for file-types like ".jpg" or ".xml".

    Args:
        input_dir: String, the directory to get the files from
        search_subfolders: Bool, whether to search in subfolders
        file_extension: String, extension of file-paths as string. e.g. ".jpg", ".xml", ".png"
        include_pattern: String, which has to be contained in a file-path
        exclude_pattern: String, which are not allowed to be contained in a file-path

    Returns: List of strings, the list of the files as strings with the complete path

    """

    file_list: List[str] = []

    if not os.path.isdir(input_dir):
        logger.warning("'{}' is not a directory".format(input_dir))
        return file_list

    if search_subfolders:
        path_generator = Path(input_dir).glob(f"**/*{file_extension}")
    else:
        path_generator = Path(input_dir).glob(f"*{file_extension}")

    file_list = [str(p) for p in path_generator]

    if include_pattern != "":
        file_list = [f for f in file_list if include_pattern in f]

    if exclude_pattern:
        file_list = [f for f in file_list if exclude_pattern not in f]

    logger.info(
        "Found '%s' files in path '%s' for files of type '%s'",
        len(file_list),
        input_dir,
        file_extension,
    )

    return file_list


def ensure_dir(file_path: str, verbose: bool = False) -> None:
    """
    Creates directories for this file-path if they do not exist.
    Does nothing for existing parts of path.

    Args:
        file_path: String, the complete path that is to be checked
        verbose: Bool, whether or not to print infos

    Returns: None

    """

    directory = os.path.dirname(file_path)

    if verbose:
        if not os.path.exists(directory) and len(directory) > 0:
            logger.warning("Create new directory: '%s' for '%s'", directory, file_path)

    os.makedirs(directory, exist_ok=True)


def get_project_path_information(
    file_path: str, dir_depth: int, code_base: str
) -> Tuple[str, str, str]:
    """
    Args:
        file_path: String, path to file
        dir_depth: String, depth of directory hierarchy
        code_base: String, the target directory beneath PROJECT_ROOT_DIR dir to change working directory to

    Returns:
        Tuple of three strings, containing directory of file_path, project root
        directory and code root directory (project_root/src)

    """
    this_dir = os.path.dirname(os.path.abspath(file_path))

    setup_path = "".join(list([".." + os.path.sep] * dir_depth))

    project_root = os.path.realpath(os.path.join(this_dir, setup_path))
    code_root = os.path.realpath(os.path.join(project_root, code_base))

    # TODO: This is unexpected given the function name, split out or change function name
    os.chdir(code_root)
    logger.info("Change working directory to: '%s'", project_root)

    # TODO: 'this_dir' is misleading, we just changed cwd
    return this_dir, project_root, code_root


def encoding_safe_imwrite(filename: str, image: np.ndarray) -> None:  # type: ignore[type-arg]
    """
    Saves (Writes) an encoded version of the given image at the given path.
    Args:
        filename: String, name of file
        image: Numpy array, image to be saved

    Returns: None

    """
    # TODO: Consider giving the (mandatory!) filetype separately to avoid having to split here
    file_type = os.path.splitext(filename)[1]
    assert file_type[0] == "."
    _, im_buf_arr = cv2.imencode(file_type, image)
    im_buf_arr.tofile(filename)


def encoding_safe_imread(filename: str) -> np.ndarray:  # type: ignore[type-arg]
    """
    Reads a image file from the given path by decoding buffered image information
    Args:
        filename: String, name of file

    Returns: Numpy array, the image at given path

    """
    image: np.ndarray = cv2.imdecode(  # type: ignore[type-arg]
        np.fromfile(filename, dtype=np.uint8), cv2.IMREAD_UNCHANGED
    )
    return image


def extract_zip_data(zip_path: str, zip_extract_dir: str, remove_existing: bool = True) -> None:
    """
    Extract the content of a zip file to a directory

    Args:
        zip_path: The path to the zip file that should be extracted
        zip_extract_dir: The path to the directory where the content should be extracted to
        remove_existing: Whether to overwrite existing data

    Returns:
        None
    """

    if os.path.isdir(zip_extract_dir):
        if remove_existing:
            logger.debug("Remove existing zip data at '%s'" % zip_extract_dir)
            shutil.rmtree(zip_extract_dir)
        else:
            logger.info("Data is already extracted, skip ...")
            return

    logger.info(f"Extract data from {zip_path} to {zip_extract_dir}")

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(zip_extract_dir)


def get_basename(file_path: str) -> str:
    """
    Get the raw basename of the given file path

    Args:
        file_path: The file from which to determine the raw basename

    Returns:
        The raw basename of the given input file_path
    """

    return os.path.splitext(os.path.basename(file_path))[0]



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for storing types that are shared across the mlcvzoo
"""

from nptyping import Int, NDArray, Shape

ImageType = NDArray[Shape["Height, Width, Any"], Int]



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

from mlcvzoo_base.configuration.structs import BaseType


class BBoxSizeTypes(BaseType):
    BBOX_ALL: str = "ALL"
    BBOX_SMALL: str = "S"
    BBOX_MEDIUM: str = "M"
    BBOX_LARGE: str = "L"


class MetricTypes(BaseType):
    GROUND_TRUTH: str = "GT"
    TRUE_POSITIVES: str = "TP"
    FALSE_POSITIVES: str = "FP"
    FALSE_NEGATIVES: str = "FN"
    RECALL: str = "RC"
    PRECISION: str = "PR"
    F1: str = "F1"
    AP: str = "AP"



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

# TODO: Consider naming this "TextPerception" instead
"""Module for handling attributes of OCR objects"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List

import numpy as np

from mlcvzoo_base.api.interfaces import Perception

DEFAULT_SCORE: float = 1.0


@dataclass
class OCRPerception(Perception):
    """Dataclass that holds attributes of OCR objects"""

    # Result of the OCR. Not all methods will recognize spaces correctly
    # so this may very well be exactly one word.
    words: List[str] = field(default_factory=list, repr=False)
    # Scores that correspond to 'predicted_words'
    word_scores: List[float] = field(default_factory=list, repr=False)
    # A chain of individual strings
    content: str = field(default="", repr=False)
    # Score of the whole prediction
    score: float = field(default=DEFAULT_SCORE, repr=False)

    def __post_init__(self) -> None:
        """
        Method to be called after instantiating an OCRPerception object.

        If the OCRModel has no text recognized return an
        OCRPerception with empty string content and empty list words.

        Returns:

        """
        if len(self.words) > 0:
            if len(self.word_scores) == 0:
                self.word_scores = [DEFAULT_SCORE for _ in self.words]
            elif self.score == DEFAULT_SCORE:
                # Convention: if not provided otherwise, the score is the mean of the word scores.
                self.score = float(np.mean(self.word_scores))
            if len(self.content) == 0:
                self.content = self.clean()
        elif len(self.content) > 0:
            self.words = [self.content]
            self.word_scores = [self.score]

    def __repr__(self) -> str:
        return f"OCRPerception({self.words} - {self.word_scores})"

    def __lt__(self, other: OCRPerception) -> bool:
        return self.score < other.score

    def to_dict(self) -> Dict[Any, Any]:
        return self.__dict__

    @staticmethod
    def from_dict(from_dict: Dict[Any, Any]) -> OCRPerception:
        """
        Instantiate a new object and update it's data
        with the information that is fed in via the
        from_dict variable

        Returns:
            the created OCRPerception object
        """

        perception = OCRPerception()

        for key in perception.__dict__.keys():
            perception.__dict__[key] = from_dict[key]

        return perception

    def to_json(self) -> Any:
        return self.to_dict()

    def clean(self, separator: str = " ") -> str:
        """
        Create a clean concatenated string that is build by the
        words of the ocr perception and a given separator

        Returns:
            the created string
        """

        return f"{separator}".join(self.words)

    def chain(self) -> str:
        """
        Convenience method to create a concatenation of words
        with an empty separator

        Returns:
            the created string
        """

        return self.clean(separator="")
